# encoding: utf-8

# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

# Initial implementation by Marcus Mönnig
#
# This Server class connects against IcingaWeb2. The monitor URL in the setup should be
# something like http://icinga2/icingaweb2
#
# Status/TODOs:
#
# * The IcingaWeb2 API is not implemented yet, so currently this implementation uses
#   two HTTP requests per action. The first fetches the HTML, then the form data is extracted and
#   then a second HTTP POST request is made which actually executed the action.
#   Once IcingaWeb2 has an API, it's probably the better choice.


from Nagstamon.servers.Generic import GenericServer
import urllib.parse
import sys
import copy
import json
import datetime
from datetime import timezone
import socket
import warnings

from bs4 import BeautifulSoup, MarkupResemblesLocatorWarning
from tzlocal import get_localzone
from requests.utils import cookiejar_from_dict
from Nagstamon.objects import (GenericHost,
                               GenericService,
                               Result)
from Nagstamon.config import conf
from Nagstamon.helpers import webbrowser_open

warnings.filterwarnings("ignore", category=MarkupResemblesLocatorWarning)

from Nagstamon.cookies import (load_cookies, cookie_data_to_jar, has_any_cookie, has_valid_cookie, delete_cookie)

def strfdelta(tdelta, fmt):
    d = {'days': tdelta.days}
    d['hours'], rem = divmod(tdelta.seconds, 3600)
    d['minutes'], d['seconds'] = divmod(rem, 60)
    return fmt.format(**d)


class IcingaDBWebServer(GenericServer):
    """
        object of Icinga server
    """
    TYPE = 'IcingaDBWeb'
    MENU_ACTIONS = ['Monitor', 'Recheck', 'Acknowledge', 'Submit check result', 'Downtime']
    STATES_MAPPING = {'hosts' : {0 : 'UP', 1 : 'DOWN', 2 : 'UNREACHABLE'},
                     'services' : {0 : 'OK', 1 : 'WARNING', 2 : 'CRITICAL', 3 : 'UNKNOWN'}}
    STATES_MAPPING_REV = {'hosts' : { 'UP': 0, 'DOWN': 1, 'UNREACHABLE': 2},
                     'services' : {'OK': 0, 'WARNING': 1, 'CRITICAL': 2, 'UNKNOWN': 3}}
    BROWSER_URLS = { 'monitor': '$MONITOR-CGI$/dashboard',
                    'hosts': '$MONITOR-CGI$/icingadb/hosts',
                    'services': '$MONITOR-CGI$/icingadb/services',
                    'history': '$MONITOR-CGI$/icingadb/history'}


    def init_config(self):
        """
            set URLs for CGI - they are static and there is no need to set them with every cycle
        """
        # dummy default empty cgi urls - get filled later when server version is known
        self.cgiurl_services = None
        self.cgiurl_hosts = None
        self.cgiurl_monitoring_health = None
        # Store current custom_filter to detect changes
        self._current_custom_filter = None

        # https://github.com/HenriWahl/Nagstamon/issues/400
        # The displayed name for host and service is the Icinga2 "internal" name and not the display_name from host/service configuration
        # This name is stored in host/service dict under key 'name' but is also used as dict key for dict containing all hosts/services
        # The "internal" name must still be used to query IcingaWeb2 and is in dict under key 'real_name' since https://github.com/HenriWahl/Nagstamon/issues/192
        self.use_display_name_host = True
        self.use_display_name_service = True
        self._cookie_cleanup_done = False

    def init_http(self):
        """
            initializing of session object
        """
        GenericServer.init_http(self)

        # Run cookie cleanup only once per application run / server instance (https://github.com/HenriWahl/Nagstamon/issues/1163)
        if not getattr(self, '_cookie_cleanup_done', False):
            self._cookie_cleanup_done = True

            # Only do this if Keycloak cookies are present at all (OIDC case) 
            if has_any_cookie(self.name, 'KEYCLOAK_SESSION') or has_any_cookie(self.name, 'KEYCLOAK_IDENTITY'):
                # If there is no valid KEYCLOAK_SESSION left, drop Icingaweb2 (session cookie persisted in DB)
                if not has_valid_cookie(self.name, 'KEYCLOAK_SESSION'):
                    deleted = delete_cookie(self.name, 'Icingaweb2')
                    if conf.debug_mode:
                        self.debug(server=self.get_name(), host='', service='',
                                debug=f'[OIDC] Cleanup-on-start: KEYCLOAK_SESSION expired/missing -> deleted {deleted}x cookie "Icingaweb2" from cookies.db')
                    cookies = load_cookies()
                    self.session.cookies = cookie_data_to_jar(self.name, cookies)


        if self.session and not 'Referer' in self.session.headers:
            self.session.headers['Referer'] = self.monitor_cgi_url

        # normally cookie auth will be used
        if not self.no_cookie_auth:
            if 'cookies' not in dir(self.session) or len(self.session.cookies) == 0:
                # get login page, thus automatically a cookie
                login = self.fetch_url('{0}/authentication/login'.format(self.monitor_url))
                if login.error == '' and login.status_code == 200:
                    form = login.result.find('form')
                    form_inputs = {}
                    for form_input in ('redirect', 'formUID', 'CSRFToken', 'btn_submit'):
                        if form is not None and not form.find('input', {'name': form_input}) is None:
                            form_inputs[form_input] = form.find('input', {'name': form_input})['value']
                        else:
                            form_inputs[form_input] = ''
                    form_inputs['username'] = self.username
                    form_inputs['password'] = self.password

                    # fire up login button with all needed data
                    self.fetch_url('{0}/authentication/login'.format(self.monitor_url), cgi_data=form_inputs)

                # Add icingaweb2-tzo cookie to ensure Icinga to correctly handle provide date time string
                utc_offset = int(get_localzone().utcoffset(datetime.datetime.now()).total_seconds())
                self.session.cookies.update(cookiejar_from_dict({'icingaweb2-tzo': f'{utc_offset}-0'}))

    def _get_status(self):
        """
            Get status from Icinga Server - only JSON
        """
        # define CGI URLs for hosts and services
        # Rebuild URLs if custom_filter has changed or URLs not yet initialized
        urls_not_initialized = (self.cgiurl_hosts is None and self.cgiurl_services is None and self.cgiurl_monitoring_health is None)
        filter_changed = (self._current_custom_filter != self.custom_filter)
        
        if urls_not_initialized or filter_changed:
            # Build optional custom filter suffixes
            # Intelligently apply filters based on their prefix
            service_filter_suffix = ''
            host_filter_suffix = ''
            
            if self.custom_filter:
                # If filter starts with 'service.' - only apply to services
                # If filter starts with 'host.' - only apply to hosts  
                # Otherwise apply to both
                if self.custom_filter.startswith('service.'):
                    service_filter_suffix = '&' + self.custom_filter
                elif self.custom_filter.startswith('host.'):
                    host_filter_suffix = '&' + self.custom_filter
                else:
                    # Generic filter - apply to both
                    service_filter_suffix = '&' + self.custom_filter
                    host_filter_suffix = '&' + self.custom_filter
            
            # Store current custom_filter to detect future changes
            self._current_custom_filter = self.custom_filter

            # services (unknown, warning or critical?)
            self.cgiurl_services = {'hard': self.monitor_cgi_url + '/icingadb/services?service.state.is_problem=y&service.state.state_type=hard&columns=service.host.name,service.host.display_name,service.name,service.display_name,service.state.hard_state,service.state.last_update,service.state.check_attempt,service.max_check_attempts,service.state.output,service.active_checks_enabled,service.notifications_enabled,service.state.is_flapping,service.state.is_acknowledged,service.state.in_downtime,service.state.last_state_change,service.state.is_reachable' + service_filter_suffix + '&format=json', \
                                    'soft': self.monitor_cgi_url + '/icingadb/services?service.state.is_problem=y&service.state.state_type=soft&columns=service.host.name,service.host.display_name,service.name,service.display_name,service.state.soft_state,service.state.last_update,service.state.check_attempt,service.max_check_attempts,service.state.output,service.active_checks_enabled,service.notifications_enabled,service.state.is_flapping,service.state.is_acknowledged,service.state.in_downtime,service.state.last_state_change,service.state.is_reachable' + service_filter_suffix + '&format=json'}
            # hosts (up or down or unreachable)
            self.cgiurl_hosts = {'hard': self.monitor_cgi_url + '/icingadb/hosts?host.state.is_problem=y&host.state.state_type=hard&columns=host.name,host.display_name,host.state.hard_state,host.state.last_update,state.check_attempt,max_check_attempts,state.output,active_checks_enabled,notifications_enabled,state.is_flapping,state.is_acknowledged,state.in_downtime,state.last_state_change' + host_filter_suffix + '&format=json', \
                                 'soft': self.monitor_cgi_url + '/icingadb/hosts?host.state.is_problem=y&host.state.state_type=soft&columns=host.name,host.display_name,host.state.soft_state,host.state.last_update,state.check_attempt,max_check_attempts,state.output,active_checks_enabled,notifications_enabled,state.is_flapping,state.is_acknowledged,state.in_downtime,state.last_state_change' + host_filter_suffix + '&format=json'}
            # monitoring health
            self.cgiurl_monitoring_health = self.monitor_cgi_url + '/health?format=json'

        # new_hosts dictionary
        self.new_hosts = dict()

        # hosts - mostly the down ones
        # now using JSON output from Icinga
        try:
            for status_type in 'hard', 'soft':
                # first attempt
                result = self.fetch_url(self.cgiurl_hosts[status_type], giveback='raw')
                # authentication errors get a status code 200 too back because its
                # HTML works fine :-(
                if result.status_code < 400 and\
                   result.result.startswith('<'):
                    # in case of auth error reset HTTP session and try again
                    self.reset_http()
                    result = self.fetch_url(self.cgiurl_hosts[status_type], giveback='raw')
                    # if it does not work again tell GUI there is a problem
                    if result.status_code < 400 and\
                       result.result.startswith('<'):
                        self.refresh_authentication = True
                        return Result(result=result.result,
                                      error='Authentication error',
                                      status_code=result.status_code)

                # purify JSON result of unnecessary control sequence \n
                jsonraw, error, status_code = copy.deepcopy(result.result.replace('\n', '')),\
                                              copy.deepcopy(result.error),\
                                              result.status_code

                # check if any error occured
                potential_error = self.check_for_error(jsonraw, error, status_code)
                if potential_error is not None:
                    return potential_error

                # Check if the backend is running
                # If it isn't running the last values stored in the database are returned/shown
                # Unfortunately we need to make a extra request for this and only, if monitoring health is possible
                if self.cgiurl_monitoring_health:
                    pass
                    # TODO: Health checks for IcingaDB and icinga-redis
                    # try:
                    #     result = self.fetch_url(self.cgiurl_monitoring_health, giveback='raw')
                    #     monitoring_health = json.loads(result.result)[0]
                    #     if (monitoring_health['is_currently_running'] == '0'):
                    #         return Result(result=monitoring_health,
                    #                       error='Icinga2 backend not running')
                    # except json.decoder.JSONDecodeError:
                    #     # https://github.com/HenriWahl/Nagstamon/issues/619
                    #     # Icinga2 monitoring health status query does not seem to work (on older version?)
                    #     self.cgiurl_monitoring_health = None

                hosts = json.loads(jsonraw)

                for host in hosts:
                    # make dict of tuples for better reading
                    h = dict(host.items())

                    # host
                    if not self.use_display_name_host:
                        # according to http://sourceforge.net/p/nagstamon/bugs/83/ it might
                        # better be name instead of display_name
                        host_name = h['name']
                    else:
                        # https://github.com/HenriWahl/Nagstamon/issues/46 on the other hand has
                        # problems with that so here we go with extra display_name option
                        host_name = h['display_name']

                    # host objects contain service objects
                    if not host_name in self.new_hosts:
                        self.new_hosts[host_name] = GenericHost()
                        self.new_hosts[host_name].name = host_name
                        self.new_hosts[host_name].server = self.name
                        self.new_hosts[host_name].status_type = status_type
                        if (status_type == 'hard'):
                            self.new_hosts[host_name].status = self.STATES_MAPPING['hosts'][int(h['state']['hard_state'])]
                        else:
                            self.new_hosts[host_name].status = self.STATES_MAPPING['hosts'][int(h['state']['soft_state'])]

                        if h['state']['last_update'].replace(".", "").isnumeric(): # new version of icingadb doesnt return unix timestamp
                            #self.new_hosts[host_name].last_check = datetime.datetime.fromtimestamp(int(float(h['state']['last_update'])))
                            utc_time = datetime.datetime.fromtimestamp(int(float(h['state']['last_update'])), tz=timezone.utc)
                        else:
                            #self.new_hosts[host_name].last_check = datetime.datetime.fromisoformat(h['state']['last_update'])
                            utc_time = datetime.datetime.fromisoformat(h['state']['last_update'])

                        local_time = utc_time.astimezone()
                        self.new_hosts[host_name].last_check = local_time.strftime("%Y-%m-%d %H:%M:%S")  # format without microseconds and tz
                        
                        self.new_hosts[host_name].attempt = "{}/{}".format(h['state']['check_attempt'],h['max_check_attempts'])
                        self.new_hosts[host_name].status_information = BeautifulSoup(str(h['state']['output']).replace('\n', ' ').strip(), 'html.parser').text
                        self.new_hosts[host_name].passiveonly = not int(h.get('active_checks_enabled') or '0')
                        self.new_hosts[host_name].notifications_disabled = not int(h.get('notifications_enabled') or '0')
                        self.new_hosts[host_name].flapping = bool(int(h['state']['is_flapping'] or 0))
                        #s['state']['is_acknowledged'] can be null, 0, 1, or 'sticky'
                        self.new_hosts[host_name].acknowledged = h['state']['is_acknowledged'] if isinstance(h['state']['is_acknowledged'], bool) else bool(int(h['state']['is_acknowledged'].replace('sticky', '1') or 0))
                        self.new_hosts[host_name].scheduled_downtime = bool(int(h['state']['in_downtime'] or 0))

                        # extra Icinga properties to solve https://github.com/HenriWahl/Nagstamon/issues/192
                        # acknowledge needs host_description and no display name
                        self.new_hosts[host_name].real_name = h['name']

                        # Icinga only updates the attempts for soft states. When hard state is reached, a flag is set and
                        # attemt is set to 1/x.
                        if (status_type == 'hard'):
                            try:
                                self.new_hosts[host_name].attempt = "{0}/{0}".format(h['max_check_attempts'])
                            except Exception:
                                self.new_hosts[host_name].attempt = "HARD"

                        # extra duration needed for calculation
                        self.new_hosts[host_name].duration = 'n/a'
                        if h['state']['last_state_change'] is not None:
                            if h['state']['last_state_change'].replace(".", "").isnumeric(): # new version of icingadb doesnt return unix timestamp
                                duration = datetime.datetime.now() - datetime.datetime.fromtimestamp(int(float(h['state']['last_state_change'])))
                            else:
                                last_state_change = datetime.datetime.fromisoformat(h['state']['last_state_change'])
                                duration = datetime.datetime.now(timezone.utc).astimezone() - last_state_change
                            
                            if duration.total_seconds() > 0:
                                self.new_hosts[host_name].duration = strfdelta(duration,'{days}d {hours}h {minutes}m {seconds}s')

                    del h, host_name
        except:
            import traceback
            traceback.print_exc(file=sys.stdout)

            # set checking flag back to False
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

        # services
        try:
            for status_type in 'hard', 'soft':
                result = self.fetch_url(self.cgiurl_services[status_type], giveback='raw')
                # purify JSON result of unnecessary control sequence \n
                jsonraw, error, status_code = copy.deepcopy(result.result.replace('\n', '')),\
                                              copy.deepcopy(result.error),\
                                              result.status_code

                if error != '' or status_code >= 400:
                    return Result(result=jsonraw,
                                  error=error,
                                  status_code=status_code)

                # check if any error occured
                self.check_for_error(jsonraw, error, status_code)

                services = copy.deepcopy(json.loads(jsonraw))

                for service in services:
                    # make dict of tuples for better reading
                    s = dict(service.items())

                    if not self.use_display_name_host:
                        # according to http://sourceforge.net/p/nagstamon/bugs/83/ it might
                        # better be name instead of display_name
                        host_name = s['host']['name']
                    else:
                        # https://github.com/HenriWahl/Nagstamon/issues/46 on the other hand has
                        # problems with that so here we go with extra display_name option
                        host_name = s['host']['display_name']

                    # host objects contain service objects
                    if not host_name in self.new_hosts:
                        self.new_hosts[host_name] = GenericHost()
                        self.new_hosts[host_name].name = host_name
                        self.new_hosts[host_name].status = 'UP'
                        # extra Icinga properties to solve https://github.com/HenriWahl/Nagstamon/issues/192
                        # acknowledge needs host_description and no display name
                        self.new_hosts[host_name].real_name = s['host']['name']

                    service_name = s['display_name']

                    # if a service does not exist create its object
                    if not service_name in self.new_hosts[host_name].services:
                        self.new_hosts[host_name].services[service_name] = GenericService()
                        self.new_hosts[host_name].services[service_name].host = host_name
                        self.new_hosts[host_name].services[service_name].name = service_name
                        self.new_hosts[host_name].services[service_name].server = self.name
                        self.new_hosts[host_name].services[service_name].status_type = status_type
                        if (status_type == 'hard'):
                            self.new_hosts[host_name].services[service_name].status = self.STATES_MAPPING['services'][int(s['state']['hard_state'])]
                        else:
                            self.new_hosts[host_name].services[service_name].status = self.STATES_MAPPING['services'][int(s['state']['soft_state'])]

                        if s['state']['last_update'].replace(".", "").isnumeric(): # new version of icingadb doesnt return unix timestamp
                            #self.new_hosts[host_name].services[service_name].last_check = datetime.datetime.fromtimestamp(int(float(s['state']['last_update'])))
                            utc_time = datetime.datetime.fromtimestamp(int(float(s['state']['last_update'])), tz=timezone.utc)
                        else:
                            #self.new_hosts[host_name].services[service_name].last_check = datetime.datetime.fromisoformat(s['state']['last_update'])
                            utc_time = datetime.datetime.fromisoformat(s['state']['last_update'])

                        local_time = utc_time.astimezone()
                        self.new_hosts[host_name].services[service_name].last_check = local_time.strftime("%Y-%m-%d %H:%M:%S")  # format without microseconds and tz

                        self.new_hosts[host_name].services[service_name].attempt = "{}/{}".format(s['state']['check_attempt'],s['max_check_attempts'])
                        self.new_hosts[host_name].services[service_name].status_information = BeautifulSoup(str(s['state']['output']).replace('\n', ' ').strip(), 'html.parser').text
                        self.new_hosts[host_name].services[service_name].passiveonly = not int(s.get('active_checks_enabled') or '0')
                        self.new_hosts[host_name].services[service_name].notifications_disabled = not int(s.get('notifications_enabled') or '0')
                        self.new_hosts[host_name].services[service_name].flapping = bool(int(s['state']['is_flapping'] or 0))
                        #s['state']['is_acknowledged'] can be null, 0, 1, or 'sticky'
                        self.new_hosts[host_name].services[service_name].acknowledged = s['state']['is_acknowledged'] if isinstance(s['state']['is_acknowledged'], bool) else bool(int(s['state']['is_acknowledged'].replace('sticky', '1') or 0))
                        self.new_hosts[host_name].services[service_name].scheduled_downtime = bool(int(s['state']['in_downtime'] or 0))
                        self.new_hosts[host_name].services[service_name].unreachable = not bool(int(s['state']['is_reachable'] or 0))

                        if self.new_hosts[host_name].services[service_name].unreachable:
                            self.new_hosts[host_name].services[service_name].status_information += " (SERVICE UNREACHABLE)"

                        # extra Icinga properties to solve https://github.com/HenriWahl/Nagstamon/issues/192
                        # acknowledge needs service_description and no display name
                        self.new_hosts[host_name].services[service_name].real_name = s['name']

                        # Icinga only updates the attempts for soft states. When hard state is reached, a flag is set and
                        # attemt is set to 1/x.
                        if (status_type == 'hard'):
                            try:
                                self.new_hosts[host_name].services[service_name].attempt = "{0}/{0}".format(s['max_check_attempts'])
                            except Exception:
                                self.new_hosts[host_name].services[service_name].attempt = "HARD"

                        # extra duration needed for calculation
                        self.new_hosts[host_name].services[service_name].duration = 'n/a'
                        if s['state']['last_state_change'] is not None:
                            if s['state']['last_update'].replace(".", "").isnumeric(): # new version of icingadb doesnt return unix timestamp
                                duration = datetime.datetime.now() - datetime.datetime.fromtimestamp(int(float(s['state']['last_state_change'])))
                            else:
                                last_state_change = datetime.datetime.fromisoformat(s['state']['last_state_change'])
                                duration = datetime.datetime.now(timezone.utc).astimezone() - last_state_change
                            
                            if duration.total_seconds() > 0:
                                self.new_hosts[host_name].services[service_name].duration = strfdelta(duration, '{days}d {hours}h {minutes}m {seconds}s')

                    del s, host_name, service_name
        except:

            import traceback
            traceback.print_exc(file=sys.stdout)

            # set checking flag back to False
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

        # some cleanup
        del jsonraw, error, hosts, services

        # dummy return in case all is OK
        return Result()


    def _set_recheck(self, host, service):
        # Set correct headers
        self.session.headers['X-Requested-With'] = 'XMLHttpRequest'
        self.session.headers.update({'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'})

        # First retrieve the info page for this host/service
        if service == '':
            url = '{0}/icingadb/host?name={1}'.format(self.monitor_cgi_url, self.hosts[host].real_name)
        else:
            url = '{0}/icingadb/service?name={1}&host.name={2}'.format(self.monitor_cgi_url,
                                                                                           self.hosts[host].services[service].real_name,
                                                                                           self.hosts[host].real_name
                                                                                           )
        result = self.fetch_url(url, giveback='raw')

        if result.error != '':
            return result
        else:
            pageraw = result.result

        pagesoup = BeautifulSoup(pageraw, 'html.parser')

        # Extract the relevant form element values
        formtag = pagesoup.select_one('form[action*="check-now"]')

        if formtag is None:
            # This typically happens when authentication expired and we got redirected HTML,
            # or the page layout changed.
            if conf.debug_mode:
                self.debug(server=self.get_name(), host=host, service=service,
                        debug=f'[Recheck] Could not find check-now form. URL={url}. Possibly not authenticated or page changed.')
            return Result(result=pageraw, error='Recheck form not found (authentication expired?)')

        if conf.debug_mode:
            self.debug(server=self.get_name(), host=host, service=service,
                       debug='[Recheck] Retrieve html form from {0}: \n{1}'.format(url,formtag.prettify()))

        btn_submit = formtag.findNext('button', {'name':'btn_submit'})['value']
        CSRFToken = formtag.findNext('input', {'name':'CSRFToken'})['value']

        # Pass these values to the check-now URL as cgi_data
        cgi_data = {}
        cgi_data['CSRFToken'] = CSRFToken
        cgi_data['btn_submit'] = btn_submit

        if service == '':
            url = '{0}/icingadb/host/check-now?name={1}'.format(self.monitor_cgi_url, self.hosts[host].real_name)
        else:
            url = '{0}/icingadb/service/check-now?name={1}&host.name={2}'.format(self.monitor_cgi_url,
                                                                                           self.hosts[host].services[service].real_name,
                                                                                           self.hosts[host].real_name
                                                                                           )
        
        response = self.fetch_url(url, giveback='raw', cgi_data=cgi_data)

        # Some debug data
        data = response.result
        error = response.error
        status_code = response.status_code
        if conf.debug_mode:
            self.debug(server=self.get_name(), host=host, service=service,
                       debug='Recheck response')
            self.debug(server=self.get_name(), host=host, service=service,
                       debug="- response: {0}".format(response))
            self.debug(server=self.get_name(), host=host, service=service,
                       debug="- data: {0}".format(data))
            self.debug(server=self.get_name(), host=host, service=service,
                       debug="- error: {0}".format(error))
            self.debug(server=self.get_name(), host=host, service=service,
                       debug="- status_code: {0}".format(status_code))


    # Overwrite function from generic server to add expire_time value
    def set_acknowledge(self, info_dict):
        '''
            different monitors might have different implementations of _set_acknowledge
        '''
        if info_dict['acknowledge_all_services'] is True:
            all_services = info_dict['all_services']
        else:
            all_services = []

        # Make sure expire_time is set
        #if not info_dict['expire_time']:
        #    info_dict['expire_time'] = None

        try:
            self._set_acknowledge(info_dict['host'],
                                  info_dict['service'],
                                  info_dict['author'],
                                  info_dict['comment'],
                                  info_dict['sticky'],
                                  info_dict['notify'],
                                  info_dict['persistent'],
                                  all_services,
                                  info_dict['expire_time'])
        except:
            import traceback
            traceback.print_exc(file=sys.stdout)
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)


    def _set_acknowledge(self, host, service, author, comment, sticky, notify, persistent, all_services=None, expire_time=None):
        # Set correct headers
        self.session.headers['X-Requested-With'] = 'XMLHttpRequest'
        self.session.headers.update({'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'})

        # First retrieve the info page for this host/service
        if service == '':
            url = '{0}/icingadb/host/acknowledge?name={1}&showCompact=1'.format(self.monitor_cgi_url,
                                                                            self.hosts[host].real_name)
        else:
            url = '{0}/icingadb/service/acknowledge?name={1}&host.name={2}&showCompact=1'.format(self.monitor_cgi_url,
                                                                                           self.hosts[host].services[service].real_name,
                                                                                           self.hosts[host].real_name
                                                                                           )

        result = self.fetch_url(url, giveback='raw')

        if result.error != '':
            return result
        else:
            pageraw = result.result

        pagesoup = BeautifulSoup(pageraw, 'html.parser')

        # Extract the relevant form element values
        formtag = pagesoup.select_one('form[action*="acknowledge"]')

        if conf.debug_mode:
            self.debug(server=self.get_name(), host=host, service=service,
                       debug='[Acknowledge] Retrieve html form from {0}: \n{1}'.format(url,formtag.prettify()))
                
        btn_submit = formtag.findNext('input', {'name':'btn_submit'})['value']
        CSRFToken = formtag.findNext('input', {'name':'CSRFToken'})['value']

        # Pass these values to the same URL as cgi_data
        cgi_data = {}
        cgi_data['CSRFToken'] = CSRFToken
        cgi_data['btn_submit'] = btn_submit
        cgi_data['comment'] = comment
        cgi_data['persistent'] = str(persistent).replace('True', 'y').replace('False', 'n')
        cgi_data['sticky'] = str(sticky).replace('True', 'y').replace('False', 'n')
        cgi_data['notify'] = str(notify).replace('True', 'y').replace('False', 'n')
        if expire_time:
            cgi_data['expire'] = 'y'
            cgi_data['expire_time'] = expire_time
        else:
            cgi_data['expire'] = 'n'

        response = self.fetch_url(url, giveback='raw', cgi_data=cgi_data)

        # Some debug data
        data = response.result
        error = response.error
        status_code = response.status_code
        if conf.debug_mode:
            self.debug(server=self.get_name(), host=host, service=service,
                       debug='Achnowledgement response')
            self.debug(server=self.get_name(), host=host, service=service,
                       debug="- response: {0}".format(response))
            self.debug(server=self.get_name(), host=host, service=service,
                       debug="- data: {0}".format(data))
            self.debug(server=self.get_name(), host=host, service=service,
                       debug="- error: {0}".format(error))
            self.debug(server=self.get_name(), host=host, service=service,
                       debug="- status_code: {0}".format(status_code))

        if len(all_services) > 0:
            for s in all_services:
                # cheap, recursive solution...
                self._set_acknowledge(host, s, author, comment, sticky, notify, persistent, [], expire_time)


    def _set_submit_check_result(self, host, service, state, comment, check_output, performance_data):
        # Set correct headers
        self.session.headers['X-Requested-With'] = 'XMLHttpRequest'
        self.session.headers.update({'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'})

        # First retrieve the info page for this host/service
        if service == '':
            url = '{0}/icingadb/host/process-checkresult?name='.format(self.monitor_cgi_url,
                                                                            self.hosts[host].real_name)
            status = self.STATES_MAPPING_REV['hosts'][state.upper()]
        else:
            url = '{0}/icingadb/service/process-checkresult?name={1}&host.name={2}'.format(self.monitor_cgi_url,
                                                                                           self.hosts[host].services[service].real_name,
                                                                                           self.hosts[host].real_name
                                                                                           )
            status = self.STATES_MAPPING_REV['services'][state.upper()]

        result = self.fetch_url(url, giveback='raw')

        if result.error != '':
            return result
        else:
            pageraw = result.result

        pagesoup = BeautifulSoup(pageraw, 'html.parser')

        # Extract the relevant form element values
        formtag = pagesoup.select_one('form[action*="process-checkresult"]')

        if conf.debug_mode:
            self.debug(server=self.get_name(), host=host, service=service,
                       debug='[Submit check result] Retrieve html form from {0}: \n{1}'.format(url,formtag.prettify()))

        btn_submit = formtag.findNext('input', {'name':'btn_submit'})['value']
        CSRFToken = formtag.findNext('input', {'name':'CSRFToken'})['value']
        
        # Pass these values to the same URL as cgi_data
        cgi_data = {}
        cgi_data['CSRFToken'] = CSRFToken
        cgi_data['btn_submit'] = btn_submit

        cgi_data['status'] = status
        cgi_data['output'] = check_output
        cgi_data['perfdata'] = performance_data

        response = self.fetch_url(url, giveback='raw', cgi_data=cgi_data)

        # Some debug data
        data = response.result
        error = response.error
        status_code = response.status_code
        if conf.debug_mode:
            self.debug(server=self.get_name(), host=host, service=service,
                       debug='Achnowledgement response')
            self.debug(server=self.get_name(), host=host, service=service,
                       debug="- response: {0}".format(response))
            self.debug(server=self.get_name(), host=host, service=service,
                       debug="- data: {0}".format(data))
            self.debug(server=self.get_name(), host=host, service=service,
                       debug="- error: {0}".format(error))
            self.debug(server=self.get_name(), host=host, service=service,
                       debug="- status_code: {0}".format(status_code))


    def _set_downtime(self, host, service, author, comment, fixed, start_time, end_time, hours, minutes):
        # Set correct headers
        self.session.headers['X-Requested-With'] = 'XMLHttpRequest'
        self.session.headers.update({'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'})

        # First retrieve the info page for this host/service
        if service == '':
            url = '{0}/icingadb/host/schedule-downtime?name={1}'.format(self.monitor_cgi_url,
                                                                            self.hosts[host].real_name)
        else:
            url = '{0}/icingadb/service/schedule-downtime?name={1}&host.name={2}&showCompact=1'.format(self.monitor_cgi_url,
                                                                                           self.hosts[host].services[service].real_name,
                                                                                           self.hosts[host].real_name
                                                                                           )
                    
        result = self.fetch_url(url, giveback='raw')

        if result.error != '':
            return result
        else:
            pageraw = result.result

        pagesoup = BeautifulSoup(pageraw, 'html.parser')

        # Extract the relevant form element values
        formtag = pagesoup.select_one('form[action*="schedule-downtime"]')

        if conf.debug_mode:
            self.debug(server=self.get_name(), host=host, service=service,
                       debug='[Set downtime] Retrieve html form from {0}: \n{1}'.format(url,formtag.prettify()))
        
        btn_submit = formtag.findNext('input', {'name':'btn_submit'})['value']
        CSRFToken = formtag.findNext('input', {'name':'CSRFToken'})['value']

        # Pass these values to the same URL as cgi_data
        cgi_data = {}
        cgi_data['CSRFToken'] = CSRFToken
        cgi_data['btn_submit'] = btn_submit
        cgi_data['comment'] = comment
        if fixed:
            cgi_data['flexible'] = 'n'
        else:
            cgi_data['flexible'] = 'y'
            cgi_data['hours'] = hours
            cgi_data['minutes'] = minutes
        if start_time == '' or start_time == 'n/a':
            start = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
        else:
            start = start_time
        if end_time == '' or end_time == 'n/a':
            end = (datetime.datetime.now() + datetime.timedelta(hours=hours, minutes=minutes)).strftime('%Y-%m-%dT%H:%M:%S')
        else:
            end = end_time

        cgi_data['start'] = start
        cgi_data['end'] = end

        response = self.fetch_url(url, giveback='raw', cgi_data=cgi_data)

        # Some debug data
        data = response.result
        error = response.error
        status_code = response.status_code
        if conf.debug_mode:
            self.debug(server=self.get_name(), host=host, service=service,
                       debug='Achnowledgement response')
            self.debug(server=self.get_name(), host=host, service=service,
                       debug="- response: {0}".format(response))
            self.debug(server=self.get_name(), host=host, service=service,
                       debug="- data: {0}".format(data))
            self.debug(server=self.get_name(), host=host, service=service,
                       debug="- error: {0}".format(error))
            self.debug(server=self.get_name(), host=host, service=service,
                       debug="- status_code: {0}".format(status_code))


    def get_start_end(self, host):
        '''
            for GUI to get actual downtime start and end from server - they may vary so it's better to get
            directly from web interface
        '''
        # Set correct headers
        self.session.headers['X-Requested-With'] = 'XMLHttpRequest'
        self.session.headers.update({'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'})

        try:
            url = '{0}/icingadb/host/schedule-downtime?name={1}'.format(self.monitor_cgi_url, self.hosts[host].real_name)
            downtime = self.fetch_url(url, giveback='raw')

            if downtime.error != '':
                return 'n/a', 'n/a'
            else:
                pageraw = downtime.result
    
            pagesoup = BeautifulSoup(pageraw, 'html.parser')

            # Super debug
            if conf.debug_mode:
                self.debug(server=self.get_name(), host=host, service='',
                           debug='[Get downtime start/end] Retrieve html from {0}: {1}'.format(url,pagesoup.prettify()))

            start = pagesoup.find('input', {'name': 'start'})['value']
            end = pagesoup.find('input', {'name': 'end'})['value']
            # give values back as tuple
            return start, end
        except:
            self.error(sys.exc_info())
            return 'n/a', 'n/a'


    def open_monitor(self, host, service=''):
        """Open monitor from tablewidget context menu."""
        if host not in self.hosts:
            print("Cannot find {}. Skipping!".format(host))
            return
        if service and service not in self.hosts[host].services:
            print("Cannot find {}::{}. Skipping!".format(host, service))
            return

        # Generate the base path for the URL
        base_path = urllib.parse.urlparse(self.monitor_url).path

        # Handle URL for host monitoring
        if service == '':
            url = '{0}/icingadb/hosts?host.state.is_problem=y&sort=host.state.severity#!{1}/icingadb/host?{2}'.format(
                self.monitor_url,
                base_path,
                urllib.parse.urlencode({'name': self.hosts[host].real_name}, quote_via=urllib.parse.quote)
            )
        else:
            # Handle URL for service monitoring
            url = '{0}/icingadb/services?service.state.is_problem=y&sort=service.state.severity%20desc#!{1}/icingadb/service?{2}'.format(
                self.monitor_url,
                base_path,
                urllib.parse.urlencode({
                    'name': self.hosts[host].services[service].real_name,
                    'host.name': self.hosts[host].real_name
                }, quote_via=urllib.parse.quote)
            )
        if conf.debug_mode:
            self.debug(server=self.get_name(), host=host, service=service,
                       debug='[Open monitor] Open host/service monitor web page {0}'.format(url))
        webbrowser_open(url)

    def get_host(self, host):
        '''
            find out ip or hostname of given host to access hosts/devices which do not appear in DNS but
            have their ip saved in Icinga
        '''
        # Host is the display name as in the GUI
        # but we need the FQDN not the display name
        host = self.hosts[host].real_name

        # the fasted method is taking hostname as used in monitor
        if conf.connect_by_host is True or host == '':
            return Result(result=host)

        # initialize ip string
        ip = ''
        address = ''

        # glue nagios cgi url and hostinfo
        cgiurl_host = self.monitor_cgi_url + '/icingadb/hosts?name={0}&columns=host.address&format=json'.format(host)

        # get host info
        hostobj = self.fetch_url(cgiurl_host, giveback='raw')
        jsonhost = hostobj.result

        try:
            # take ip from json output
            result = json.loads(jsonhost)[0]
            ip = result["address"]

            # print IP in debug mode
            if conf.debug_mode is True:
                self.debug(server=self.get_name(), host=host, debug='IP of %s:' % (host) + ' ' + ip)

            # when connection by DNS is not configured do it by IP
            if conf.connect_by_dns is True:
                # try to get DNS name for ip, if not available use ip
                try:
                    address = socket.gethostbyaddr(ip)[0]
                except socket.error:
                    address = ip
            else:
                address = ip
        except Exception:
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

        # do some cleanup
        del hostobj

        # give back host or ip
        return Result(result=address)



# Xlib.xobject.icccm -- ICCCM structures
#
#    Copyright (C) 2000 Peter Liljenberg <petli@ctrl-c.liu.se>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,  USA

from Xlib import X, Xutil
from Xlib.protocol import rq

Aspect = rq.Struct( rq.Int32('num'), rq.Int32('denum') )

WMNormalHints = rq.Struct( rq.Card32('flags'),
                           rq.Pad(16),
                           rq.Int32('min_width', default = 0),
                           rq.Int32('min_height', default = 0),
                           rq.Int32('max_width', default = 0),
                           rq.Int32('max_height', default = 0),
                           rq.Int32('width_inc', default = 0),
                           rq.Int32('height_inc', default = 0),
                           rq.Object('min_aspect', Aspect, default = (0, 0)),
                           rq.Object('max_aspect', Aspect, default = (0, 0)),
                           rq.Int32('base_width', default = 0),
                           rq.Int32('base_height', default = 0),
                           rq.Int32('win_gravity', default = 0),
                           )

WMHints = rq.Struct( rq.Card32('flags'),
                     rq.Card32('input', default = 0),
                     rq.Set('initial_state', 4,
                            # withdrawn is totally bogus according to
                            # ICCCM, but some window managers seem to
                            # use this value to identify dockapps.
                            # Oh well.
                            ( Xutil.WithdrawnState,
                              Xutil.NormalState,
                              Xutil.IconicState ),
                            default = Xutil.NormalState),
                     rq.Pixmap('icon_pixmap', default = 0),
                     rq.Window('icon_window', default = 0),
                     rq.Int32('icon_x', default = 0),
                     rq.Int32('icon_y', default = 0),
                     rq.Pixmap('icon_mask', default = 0),
                     rq.Window('window_group', default = 0),
                     )

WMState = rq.Struct( rq.Set('state', 4,
                            ( Xutil.WithdrawnState,
                              Xutil.NormalState,
                              Xutil.IconicState )),
                     rq.Window('icon', ( X.NONE, )),
                     )


WMIconSize = rq.Struct( rq.Card32('min_width'),
                        rq.Card32('min_height'),
                        rq.Card32('max_width'),
                        rq.Card32('max_height'),
                        rq.Card32('width_inc'),
                        rq.Card32('height_inc'),
                        )



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
from Nagstamon.config import (conf,
                              OS,
                              OS_MACOS)
from Nagstamon.qui.constants import (HEADERS,
                                     HEADERS_KEYS_COLUMNS,
                                     HEADERS_HEADERS_COLUMNS,
                                     SORT_ORDER,
                                     SPACE)
from Nagstamon.qui.qt import (QPushButton,
                              QSizePolicy,
                              Qt,
                              QVBoxLayout,
                              Signal,
                              Slot)
from Nagstamon.qui.widgets.buttons import (Button,
                                           PushButtonBrowserURL)
from Nagstamon.qui.widgets.labels import (ClosingLabel,
                                          ServerStatusLabel)
from Nagstamon.qui.widgets.layout import HBoxLayout
from Nagstamon.qui.widgets.treeview import TreeView


class ServerVBox(QVBoxLayout):
    """
    one VBox per server containing buttons and hosts/services listview
    """
    # used to update status label text like 'Connected-'
    change_label_status = Signal(str, str)

    # signal to submit server to authentication dialog for credentials
    authenticate_credentials = Signal(str)

    # signal to submit server to weblogin dialog as browser
    authenticate_weblogin = Signal(str)

    # handle TLS error button
    button_fix_tls_error_show = Signal()
    button_fix_tls_error_hide = Signal()

    # buttons pressed, which need dialogs
    button_edit_pressed = Signal(str)
    button_fix_tls_error_pressed = Signal(str, bool)

    # open dialog - may need closing the statuswindow
    open_dialog = Signal()

    def __init__(self, server, parent=None):
        QVBoxLayout.__init__(self, parent)

        self.parent_statuswindow = parent

        # no space around
        self.setSpacing(0)
        self.setContentsMargins(0, 0, 0, 0)

        # server the vbox belongs to
        self.server = server

        # header containing monitor name, buttons and status
        self.header = HBoxLayout(spacing=SPACE, parent=parent)
        self.addLayout(self.header)
        # top and bottom should be kept by padding
        self.header.setContentsMargins(0, 0, SPACE, 0)

        self.label = ClosingLabel(parent=parent)
        self.label.mouse_released.connect(self.parent_statuswindow.hide_window)
        self.update_label()
        self.button_monitor = PushButtonBrowserURL(text='Monitor',
                                                   parent=parent,
                                                   server=self.server,
                                                   url_type='monitor')
        self.button_monitor.webbrowser_opened.connect(self.parent_statuswindow.hide_window)
        self.button_hosts = PushButtonBrowserURL(text='Hosts',
                                                 parent=parent,
                                                 server=self.server,
                                                 url_type='hosts')
        self.button_hosts.webbrowser_opened.connect(self.parent_statuswindow.hide_window)
        self.button_services = PushButtonBrowserURL(text='Services',
                                                    parent=parent,
                                                    server=self.server,
                                                    url_type='services')
        self.button_services.webbrowser_opened.connect(self.parent_statuswindow.hide_window)
        self.button_history = PushButtonBrowserURL(text='History',
                                                   parent=parent,
                                                   server=self.server,
                                                   url_type='history')
        self.button_history.webbrowser_opened.connect(self.parent_statuswindow.hide_window)
        self.button_edit = Button('Edit',
                                  parent=parent)

        # use label instead of spacer to be clickable
        self.label_stretcher = ClosingLabel('', parent=parent)
        self.label_stretcher.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Expanding)
        self.label_stretcher.mouse_released.connect(self.parent_statuswindow.hide_window)

        self.label_status = ServerStatusLabel(parent=parent)
        self.label_status.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)

        self.button_authenticate = QPushButton('Authenticate', parent=parent)

        self.button_fix_tls_error = QPushButton('Fix error', parent=parent)

        # avoid useless spaces in macOS when server has nothing to show
        # see https://bugreports.qt.io/browse/QTBUG-2699
        self.button_monitor.setAttribute(Qt.WidgetAttribute.WA_LayoutUsesWidgetRect)
        self.button_history.setAttribute(Qt.WidgetAttribute.WA_LayoutUsesWidgetRect)
        self.button_services.setAttribute(Qt.WidgetAttribute.WA_LayoutUsesWidgetRect)
        self.button_hosts.setAttribute(Qt.WidgetAttribute.WA_LayoutUsesWidgetRect)
        self.button_edit.setAttribute(Qt.WidgetAttribute.WA_LayoutUsesWidgetRect)
        self.button_authenticate.setAttribute(Qt.WidgetAttribute.WA_LayoutUsesWidgetRect)
        self.button_fix_tls_error.setAttribute(Qt.WidgetAttribute.WA_LayoutUsesWidgetRect)

        self.button_monitor.clicked.connect(self.button_monitor.open_url)
        self.button_hosts.clicked.connect(self.button_hosts.open_url)
        self.button_services.clicked.connect(self.button_services.open_url)
        self.button_history.clicked.connect(self.button_history.open_url)
        self.button_edit.clicked.connect(self.edit_server)

        self.header.addWidget(self.label)
        self.header.addWidget(self.button_monitor)
        self.header.addWidget(self.button_hosts)
        self.header.addWidget(self.button_services)
        self.header.addWidget(self.button_history)
        self.header.addWidget(self.button_edit)

        self.header.addWidget(self.label_stretcher)

        self.header.addWidget(self.label_status)
        self.header.addWidget(self.button_authenticate)
        self.header.addWidget(self.button_fix_tls_error)

        self.open_dialog.connect(self.parent_statuswindow.hide_window)

        # attempt to get header strings
        try:
            # when stored as simple lowercase keys
            sort_column = HEADERS_KEYS_COLUMNS[conf.default_sort_field]
        except Exception:
            # when as legacy stored as presentation string
            sort_column = HEADERS_HEADERS_COLUMNS[conf.default_sort_field]

        # convert sort order to number as used in Qt.SortOrder
        sort_order = SORT_ORDER[conf.default_sort_order.lower()]

        self.table = TreeView(len(HEADERS) + 1, 0, sort_column, sort_order, self.server, parent=parent)

        # delete vbox if thread quits
        self.table.worker_thread.finished.connect(self.delete)

        # connect worker to status label to reflect connectivity
        self.table.worker.change_label_status.connect(self.label_status.change)
        self.table.worker.restore_label_status.connect(self.label_status.restore)
        self.table.worker.authentication_needed.connect(self.parent_statuswindow.show_window_for_authentication)

        # care about authentications
        self.button_authenticate.clicked.connect(self.authenticate_server)
        # somehow a long way to connect the signal with the slot but works
        self.authenticate_credentials.connect(self.parent_statuswindow.injected_dialogs.authentication.show_auth_dialog)
        self.authenticate_weblogin.connect(self.parent_statuswindow.injected_dialogs.weblogin.show_browser)
        self.parent_statuswindow.injected_dialogs.authentication.update.connect(self.update_label)

        # start ignoring TLS trouble when button clicked
        self.button_fix_tls_error.clicked.connect(self.fix_tls_error)

        # connect button signals to dialogs
        self.button_edit_pressed.connect(self.parent_statuswindow.injected_dialogs.server.edit)
        self.button_fix_tls_error_pressed.connect(self.parent_statuswindow.injected_dialogs.server.edit)

        self.addWidget(self.table, 1)

        # as default do not show anything
        self.show_only_header()

    def get_real_height(self):
        """
        return summarized real height of hbox items and table
        """
        height = self.table.get_real_height()
        if self.label.isVisible() and self.button_monitor.isVisible():
            # compare item heights, decide to take the largest and add 2 time the MARGIN (top and bottom)
            if self.label.sizeHint().height() > self.button_monitor.sizeHint().height():
                height += self.label.sizeHint().height() + 2
            else:
                height += self.button_monitor.sizeHint().height() + 2
        return height

    @Slot()
    def show_all(self):
        """
        show all items in server vbox
        """
        self.button_authenticate.hide()
        self.button_edit.show()
        self.button_fix_tls_error.hide()
        self.button_history.show()
        self.button_hosts.show()
        self.button_monitor.show()
        self.button_services.show()
        self.label.show()
        self.label_status.show()
        self.label_stretcher.show()
        # special table treatment
        self.table.show()
        # self.table.is_shown = True

    @Slot()
    def show_only_header(self):
        """
        show all items in server vbox except the table - not needed if empty or major connection problem
        """
        self.button_authenticate.hide()
        self.button_edit.show()
        self.button_history.show()
        self.button_hosts.show()
        self.button_fix_tls_error.hide()
        self.button_monitor.show()
        self.button_services.show()
        self.label.show()
        self.label_status.show()
        self.label_stretcher.show()
        # special table treatment
        self.table.hide()

    @Slot()
    def hide_all(self):
        """
        hide all items in server vbox
        """
        self.button_authenticate.hide()
        self.button_edit.hide()
        self.button_fix_tls_error.hide()
        self.button_history.hide()
        self.button_hosts.hide()
        self.button_monitor.hide()
        self.button_services.hide()
        self.label.hide()
        self.label_status.hide()
        self.label_stretcher.hide()
        # special table treatment
        self.table.hide()

    @Slot()
    def delete(self):
        """
        delete VBox and its children
        """
        for widget in (self.label,
                       self.button_monitor,
                       self.button_hosts,
                       self.button_services,
                       self.button_history,
                       self.button_edit,
                       self.label_status,
                       self.label_stretcher,
                       self.button_authenticate,
                       self.button_fix_tls_error):
            widget.hide()
            widget.deleteLater()
        self.removeItem(self.header)
        self.header.deleteLater()
        self.table.hide()
        self.table.deleteLater()
        self.deleteLater()

    def edit_server(self):
        """
        call dialogs.server.edit() with server name
        """
        if not conf.fullscreen and not conf.windowed:
            self.open_dialog.emit()
        self.button_edit_pressed.emit(self.server.name)

    def authenticate_server(self):
        """
        send signal to open authentication dialog with self.server.name
        """
        if self.server.authentication != 'web':
            self.authenticate_credentials.emit(self.server.name)
        else:
            self.authenticate_weblogin.emit(self.server.name)

    @Slot()
    def update_label(self):
        if self.server.authentication != 'web':
            self.label.setText(f'<big><b>&nbsp;{self.server.username}@{self.server.name}</b></big>')
        else:
            self.label.setText(f'<big><b>&nbsp;{self.server.name}</b></big>')
        # let label padding keep top and bottom space - apparently not necessary on OSX
        if OS != OS_MACOS:
            self.label.setStyleSheet('''padding-top: {0}px;
                                        padding-bottom: {0}px;'''.format(SPACE))

    @Slot()
    def fix_tls_error(self):
        """
        call dialogs.server.edit() with server name and showing extra options
        """
        if not conf.fullscreen and not conf.windowed:
            self.open_dialog.emit()
        self.button_fix_tls_error_pressed.emit(self.server.name, True)


# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from copy import deepcopy
from functools import wraps
import os
from urllib.parse import quote

from Nagstamon.config import conf, CONFIG_STRINGS, BOOLPOOL, Server
from Nagstamon.cookies import delete_cookie
from Nagstamon.qui.globals import (ecp_available,
                                   kerberos_available)
from Nagstamon.qui.qt import (QFileDialog,
                              QMessageBox,
                              QStyle,
                              QWidget,
                              Signal,
                              Slot)
from Nagstamon.qui.dialogs.dialog import Dialog
from Nagstamon.servers import (create_server,
                               servers,
                               SERVER_TYPES)


class DialogServer(Dialog):
    """
    dialog used to set up one single server
    """

    # tell server has been edited
    edited = Signal()

    # signal to emit when ok button is pressed - used to remove previous server
    edited_remove_previous = Signal(str)

    # signal to emit when ok button is pressed - used to update the list of servers
    edited_update_list = Signal(str, str, str)

    # signal to emit when a new server vbox has to be created
    create_server_vbox = Signal(str)

    # signal to emit when web cookies should be deleted for a server
    delete_web_cookies = Signal(str, QWidget)

    def __init__(self):
        Dialog.__init__(self, 'settings_server')
        # file chooser Dialog
        self.file_chooser = QFileDialog()
        # configuration for server
        self.server_conf = None
        # define checkbox-to-widgets dependencies which apply at initialization
        # which widgets have to be hidden because of irrelevance
        # dictionary holds checkbox/radiobutton as key and relevant widgets in a list
        self.TOGGLE_DEPS = {
            self.window.input_checkbox_use_autologin: [self.window.label_autologin_key,
                                                       self.window.input_lineedit_autologin_key],
            self.window.input_checkbox_use_proxy: [self.window.groupbox_proxy],

            self.window.input_checkbox_use_proxy_from_os: [self.window.label_proxy_address,
                                                           self.window.input_lineedit_proxy_address,
                                                           self.window.label_proxy_username,
                                                           self.window.input_lineedit_proxy_username,
                                                           self.window.label_proxy_password,
                                                           self.window.input_lineedit_proxy_password],
            self.window.input_checkbox_show_options: [self.window.groupbox_options],
            self.window.input_checkbox_custom_cert_use: [self.window.label_custom_ca_file,
                                                         self.window.input_lineedit_custom_cert_ca_file,
                                                         self.window.button_choose_custom_cert_ca_file]}

        self.TOGGLE_DEPS_INVERTED = [self.window.input_checkbox_use_proxy_from_os]

        # these widgets are shown or hidden depending on server type properties
        # the servers listed at each widget do need them
        self.VOLATILE_WIDGETS = {
            self.window.label_monitor_cgi_url: ['Nagios', 'Icinga', 'Thruk', 'Sensu', 'SensuGo'],
            self.window.input_lineedit_monitor_cgi_url: ['Nagios', 'Icinga', 'Thruk', 'Sensu', 'SensuGo'],
            self.window.input_checkbox_use_autologin: ['Centreon', 'monitos4x', 'Thruk'],
            self.window.input_lineedit_autologin_key: ['Centreon', 'monitos4x', 'Thruk'],
            self.window.label_autologin_key: ['Centreon', 'monitos4x', 'Thruk'],
            self.window.input_checkbox_no_cookie_auth: ['IcingaWeb2', 'Sensu'],
            self.window.input_checkbox_use_display_name_host: ['Icinga', 'IcingaWeb2'],
            self.window.input_checkbox_use_display_name_service: ['Icinga', 'IcingaWeb2', 'Thruk'],
            self.window.input_checkbox_use_description_name_service: ['Zabbix'],
            self.window.input_checkbox_force_authuser: ['Checkmk Multisite'],
            self.window.groupbox_checkmk_views: ['Checkmk Multisite'],
            self.window.input_lineedit_host_filter: ['op5Monitor'],
            self.window.input_lineedit_service_filter: ['op5Monitor'],
            self.window.label_service_filter: ['op5Monitor'],
            self.window.label_host_filter: ['op5Monitor'],
            self.window.input_lineedit_hashtag_filter: ['Opsview'],
            self.window.label_hashtag_filter: ['Opsview'],
            self.window.input_checkbox_can_change_only: ['Opsview'],
            self.window.label_monitor_site: ['Sensu'],
            self.window.input_lineedit_monitor_site: ['Sensu'],
            self.window.label_map_to_hostname: ['Prometheus', 'Alertmanager'],
            self.window.input_lineedit_map_to_hostname: ['Prometheus', 'Alertmanager'],
            self.window.input_checkbox_treat_services_as_alerts: ['LibreNMS'],
            self.window.label_map_to_servicename: ['Prometheus', 'Alertmanager'],
            self.window.input_lineedit_map_to_servicename: ['Prometheus', 'Alertmanager'],
            self.window.label_map_to_status_information: ['Prometheus', 'Alertmanager'],
            self.window.input_lineedit_map_to_status_information: ['Prometheus', 'Alertmanager'],
            self.window.label_alertmanager_filter: ['Alertmanager'],
            self.window.input_lineedit_alertmanager_filter: ['Alertmanager'],
            self.window.label_map_to_ok: ['Alertmanager'],
            self.window.input_lineedit_map_to_ok: ['Alertmanager'],
            self.window.label_map_to_unknown: ['Alertmanager'],
            self.window.input_lineedit_map_to_unknown: ['Alertmanager'],
            self.window.label_map_to_warning: ['Alertmanager'],
            self.window.input_lineedit_map_to_warning: ['Alertmanager'],
            self.window.label_map_to_critical: ['Alertmanager'],
            self.window.input_lineedit_map_to_critical: ['Alertmanager'],
            self.window.label_map_to_down: ['Alertmanager'],
            self.window.input_lineedit_map_to_down: ['Alertmanager'],
            self.window.input_lineedit_notification_filter: ['IcingaDBWebNotifications'],
            self.window.label_notification_filter: ['IcingaDBWebNotifications'],
            self.window.input_lineedit_notification_lookback: ['IcingaDBWebNotifications'],
            self.window.label_notification_lookback: ['IcingaDBWebNotifications'],
            self.window.input_lineedit_custom_filter: ['IcingaDBWeb'],
            self.window.label_custom_filter: ['IcingaDBWeb'],
            self.window.label_disabled_backends: ['Thruk'],
            self.window.input_lineedit_disabled_backends: ['Thruk'],
        }

        # to be used when selecting authentication method Kerberos or Web
        self.AUTHENTICATION_WIDGETS = [
            self.window.label_username,
            self.window.input_lineedit_username,
            self.window.label_password,
            self.window.input_lineedit_password,
            self.window.input_checkbox_save_password]

        self.AUTHENTICATION_BEARER_WIDGETS = [
            self.window.label_username,
            self.window.input_lineedit_username]

        self.AUTHENTICATION_ECP_WIDGETS = [
            self.window.label_idp_ecp_endpoint,
            self.window.input_lineedit_idp_ecp_endpoint]

        # custom CA is not possible with authentification method Web due to the underlying Qt WebEngine aka Chromium
        self.CUSTOM_CERT_CA_WIDGETS = [ self.window.input_checkbox_custom_cert_use ]

        # fill default order fields combobox with monitor server types
        self.window.input_combobox_type.addItems(sorted(SERVER_TYPES.keys(), key=str.lower))
        # default to Nagios as it is the mostly used monitor server
        self.window.input_combobox_type.setCurrentText('Nagios')

        # set folder and play symbols to choose and play buttons
        self.window.button_choose_custom_cert_ca_file.setText('')
        self.window.button_choose_custom_cert_ca_file.setIcon(
            self.window.button_choose_custom_cert_ca_file.style().standardIcon(QStyle.StandardPixmap.SP_DirIcon))
        # connect choose custom cert CA file button with file dialog
        self.window.button_choose_custom_cert_ca_file.clicked.connect(self.choose_custom_cert_ca_file)

        # fill authentication combobox
        combobox_authentication_items = ['Basic', 'Bearer', 'Digest']
        if ecp_available:
            combobox_authentication_items.append('ECP')
        if kerberos_available:
            combobox_authentication_items.append('Kerberos')
        # Keyring availability is checked in Config - only show Web authentication if keyring is available,
        # because otherwise encryption key cannot be stored securely
        if conf.is_keyring_available():
            combobox_authentication_items.append('Web')
        # sort items
        self.window.input_combobox_authentication.addItems(sorted(combobox_authentication_items))

        # detect change of a server type which leads to certain options shown or hidden
        self.window.input_combobox_type.activated.connect(self.toggle_type)

        # when authentication is changed to Kerberos then disable username/password as they are now useless
        self.window.input_combobox_authentication.activated.connect(self.toggle_authentication)

        # reset Checkmk views
        self.window.button_checkmk_view_hosts_reset.clicked.connect(self.checkmk_view_hosts_reset)
        self.window.button_checkmk_view_services_reset.clicked.connect(self.checkmk_view_services_reset)

        self.window.button_delete_web_cookies.clicked.connect(self.on_delete_web_cookies)

        # mode needed for evaluate dialog after ok button pressed - defaults to 'new'
        self.mode = 'new'

    @Slot(int)
    def toggle_type(self, server_type_index=0):
        # server_type_index is not needed - we get the server type from .currentText()
        # check if server type is listed in volatile widgets to decide if it has to be shown or hidden
        for widget, server_types in self.VOLATILE_WIDGETS.items():
            if self.window.input_combobox_type.currentText() in server_types:
                widget.show()
            else:
                widget.hide()

    @Slot()
    def toggle_authentication(self):
        """
        when authentication is changed to Kerberos then disable username/password as they are now useless
        """
        if self.window.input_combobox_authentication.currentText() == 'Kerberos':
            for widget in self.AUTHENTICATION_WIDGETS:
                widget.hide()
        else:
            for widget in self.AUTHENTICATION_WIDGETS:
                widget.show()

        if self.window.input_combobox_authentication.currentText() == 'ECP':
            for widget in self.AUTHENTICATION_ECP_WIDGETS:
                widget.show()
        else:
            for widget in self.AUTHENTICATION_ECP_WIDGETS:
                widget.hide()

        # change credential input for bearer auth
        if self.window.input_combobox_authentication.currentText() == 'Bearer':
            for widget in self.AUTHENTICATION_BEARER_WIDGETS:
                widget.hide()
                self.window.label_password.setText('Token')
        else:
            for widget in self.AUTHENTICATION_BEARER_WIDGETS:
                widget.show()
                self.window.label_password.setText('Password')

        # no need for username + password when using Web authentication
        if self.window.input_combobox_authentication.currentText() == 'Web':
            for widget in self.AUTHENTICATION_WIDGETS + self.CUSTOM_CERT_CA_WIDGETS:
                widget.hide()
            self.window.button_delete_web_cookies.show()
        else:
            for widget in self.AUTHENTICATION_WIDGETS + self.CUSTOM_CERT_CA_WIDGETS:
                widget.show()
            self.window.button_delete_web_cookies.hide()

        # after hiding authentication widgets dialog might shrink
        self.window.adjustSize()

    def dialog_decoration(method, *args, **kwargs):
        """
        try with a decorator instead of repeated calls
        """

        # the function which decorates method
         # wraps is used to keep the original method's name and docstring
        @wraps(method)
        def decoration_function(self, *args, **kwargs):
            """
                self.server_conf has to be set by decorated method
            """
            # previous server conf only useful when editing - defaults to None
            self.previous_server_conf = None

            # call decorated method
            method(self, *args, **kwargs)

            # run through all input widgets and apply defaults from config
            for widget in self.window.__dict__:

                if widget.startswith('input_'):
                    if widget.startswith('input_checkbox_'):
                        setting = widget.split('input_checkbox_')[1]
                        self.window.__dict__[widget].setChecked(self.server_conf.__dict__[setting])
                    elif widget.startswith('input_radiobutton_'):
                        setting = widget.split('input_radiobutton_')[1]
                        self.window.__dict__[widget].setChecked(self.server_conf.__dict__[setting])
                    elif widget.startswith('input_combobox_'):
                        setting = widget.split('input_combobox_')[1]
                        self.window.__dict__[widget].setCurrentText(self.server_conf.__dict__[setting])
                    elif widget.startswith('input_lineedit_'):
                        setting = widget.split('input_lineedit_')[1]
                        self.window.__dict__[widget].setText(self.server_conf.__dict__[setting])
                    elif widget.startswith('input_spinbox_'):
                        setting = widget.split('input_spinbox_')[1]
                        self.window.__dict__[widget].setValue(self.server_conf.__dict__[setting])

            # set the current authentication type by using capitalized first letter via .title()
            self.window.input_combobox_authentication.setCurrentText(self.server_conf.authentication.title())

            # initially hide unnecessary widgets
            self.toggle_type()

            # disable unneeded authentication widgets if Kerberos is used
            self.toggle_authentication()

            # apply toggle-dependencies between checkboxes and certain widgets
            self.toggle_toggles()

            # open extra options if wanted, for example, by button_fix_tls_error
            if 'show_options' in self.__dict__:
                if self.show_options:
                    self.window.input_checkbox_show_options.setChecked(True)

            # important final size adjustment
            self.window.adjustSize()

            # if running on macOS with disabled dock icon, the dock icon might have to be made visible
            # to make Nagstamon accept keyboard input
            self.check_macos_dock_icon_fix_show.emit()

            # force modality - using .open() still allows to switch to main settings dialog
            self.window.exec()

            # en reverse the dock icon might be hidden again after a potential keyboard input
            self.check_macos_dock_icon_fix_hide.emit()

        # give back decorated function
        return decoration_function

    @Slot()
    @dialog_decoration
    def new(self):
        """
        create new server, set default values
        """
        self.mode = 'new'

        # create a new server config object
        self.server_conf = Server()
        # window title might be pretty simple
        self.window.setWindowTitle('New server')

    @Slot(str)
    @dialog_decoration
    def edit(self, name=None, show_options=False):
        """
        edit existing server
        when called by Edit button in ServerVBox use given server name to get server config
        """

        self.mode = 'edit'
        # shorter server conf
        # if name is None:
        #     self.server_conf = conf.servers[dialogs.settings.window.list_servers.currentItem().text()]
        # else:
        #     self.server_conf = conf.servers[name]
        self.server_conf = conf.servers[name]
        # store monitor name in case it will be changed
        self.previous_server_conf = deepcopy(self.server_conf)
        # set window title
        self.window.setWindowTitle('Edit %s' % (self.server_conf.name))
        # set self.show_options to give value to decorator
        self.show_options = show_options

    @Slot(str)
    @dialog_decoration
    def copy(self, name=None):
        """
        copy existing server
        """
        self.mode = 'copy'
        # shorter server conf
        self.server_conf = deepcopy(conf.servers[name])
        # set window title before name change to reflect copy
        self.window.setWindowTitle(f'Copy {self.server_conf.name}')
        # indicate copy of another server
        self.server_conf.name = f'Copy of {self.server_conf.name}'

    def ok(self):
        """
        evaluate the state of widgets to get new configuration
        """
        # strip name to avoid whitespace
        server_name = self.window.input_lineedit_name.text().strip()

        # check that no duplicate name exists
        if server_name in conf.servers and \
                (self.mode in ['new', 'copy'] or
                 self.mode == 'edit' and self.server_conf != conf.servers[server_name]):
            # cry if duplicate name exists
            QMessageBox.critical(self.window,
                                 'Nagstamon',
                                 f'The monitor server name <b>{server_name}</b> is already used.',
                                 QMessageBox.StandardButton.Ok)
        else:
            # get configuration from UI
            for widget in self.window.__dict__:
                if widget.startswith('input_'):
                    if widget.startswith('input_checkbox_'):
                        setting = widget.split('input_checkbox_')[1]
                        self.server_conf.__dict__[setting] = self.window.__dict__[widget].isChecked()
                    elif widget.startswith('input_radiobutton_'):
                        setting = widget.split('input_radiobutton_')[1]
                        self.server_conf.__dict__[setting] = self.window.__dict__[widget].isChecked()
                    elif widget.startswith('input_combobox_'):
                        setting = widget.split('input_combobox_')[1]
                        self.server_conf.__dict__[setting] = self.window.__dict__[widget].currentText()
                    elif widget.startswith('input_lineedit_'):
                        setting = widget.split('input_lineedit_')[1]
                        self.server_conf.__dict__[setting] = self.window.__dict__[widget].text()
                    elif widget.startswith('input_spinbox_'):
                        setting = widget.split('input_spinbox_')[1]
                        self.server_conf.__dict__[setting] = self.window.__dict__[widget].value()

            # URLs should not end with / - clean it
            self.server_conf.monitor_url = self.server_conf.monitor_url.rstrip('/')
            self.server_conf.monitor_cgi_url = self.server_conf.monitor_cgi_url.rstrip('/')

            # convert some strings to integers and bools
            for item in self.server_conf.__dict__:
                if type(self.server_conf.__dict__[item]) == str:
                    # when an item is not one of those which always have to be strings, then it might be OK to convert it
                    if not item in CONFIG_STRINGS:
                        if self.server_conf.__dict__[item] in BOOLPOOL:
                            self.server_conf.__dict__[item] = BOOLPOOL[self.server_conf.__dict__[item]]
                        elif self.server_conf.__dict__[item].isdecimal():
                            self.server_conf.__dict__[item] = int(self.server_conf.__dict__[item])

            # store lowered authentication type
            self.server_conf.authentication = self.server_conf.authentication.lower()

            # edited servers will be deleted and recreated with new configuration
            if self.mode == 'edit':
                # remove old server vbox from the status window if still running
                self.edited_remove_previous.emit(self.previous_server_conf.name)

                # delete previous name
                conf.servers.pop(self.previous_server_conf.name)

                # delete edited and now not needed server instance - if it exists
                if self.previous_server_conf.name in servers.keys():
                    servers.pop(self.previous_server_conf.name)

            # some monitor servers do not need cgi-url - reuse self.VOLATILE_WIDGETS to find out which one
            if self.server_conf.type not in self.VOLATILE_WIDGETS[self.window.input_lineedit_monitor_cgi_url]:
                self.server_conf.monitor_cgi_url = self.server_conf.monitor_url

            # add new server configuration in every case and use stripped name to avoid spaces
            self.server_conf.name = server_name
            conf.servers[server_name] = self.server_conf

            # add new server instance to global servers dict
            servers[server_name] = create_server(self.server_conf)
            if self.server_conf.enabled:
                servers[server_name].enabled = True
                # create vbox
                self.create_server_vbox.emit(server_name)

            # reorder servers in dict to reflect changes
            servers_freshly_sorted = sorted(servers.items())
            servers.clear()
            servers.update(servers_freshly_sorted)
            del servers_freshly_sorted

            # refresh the list of servers, give call the current server name to highlight it
            self.edited_update_list.emit('list_servers', 'servers', self.server_conf.name)

            # tell the main window about changes (Zabbix, Opsview, for example)
            self.edited.emit()

            # delete the old server .conf file to reflect name changes
            # new one will be written soon
            if self.previous_server_conf is not None:
                conf.delete_file('servers', f"server_{quote(self.previous_server_conf.name, safe='')}.conf")

            # store server settings
            conf.save_multiple_config('servers', 'server')

        # call close and macOS dock icon treatment from ancestor
        super().ok()

    @Slot()
    def choose_custom_cert_ca_file(self):
        """
        show dialog for selection of non-default browser
        """
        file_filter = 'All files (*)'
        file = self.file_chooser.getOpenFileName(self.window,
                                                    directory=os.path.expanduser('~'),
                                                    filter=file_filter)[0]

        # only take filename if QFileDialog gave something useful back
        if file != '':
            self.window.input_lineedit_custom_cert_ca_file.setText(file)

    @Slot()
    def checkmk_view_hosts_reset(self):
        self.window.input_lineedit_checkmk_view_hosts.setText('nagstamon_hosts')

    @Slot()
    def checkmk_view_services_reset(self):
        self.window.input_lineedit_checkmk_view_services.setText('nagstamon_svc')

    def on_delete_web_cookies(self):
        """
        delete all cookies for given server
        """
        # strip name to avoid whitespace
        server_name = self.window.input_lineedit_name.text().strip()
        self.delete_web_cookies.emit(server_name, self)

    @Slot(str)
    def delete_web_cookies_action(self, server_name=None, parent=None):
        # strip name to avoid whitespace
        if server_name is None:
            server_name = self.window.input_lineedit_name.text().strip()
        if parent is None:
            parent = self.window

        reply = QMessageBox.question(parent, 'Nagstamon',
                                     f'Do you really want to delete <b>all web cookies</b> of monitor server <b>{server_name}</b>?',
                                     QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                                     QMessageBox.StandardButton.No)
        if reply == QMessageBox.StandardButton.Yes:
            delete_cookie(server_name)



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from Nagstamon.config import (conf,
                              OS,
                              OS_MACOS)
from Nagstamon.qui.dialogs.about import DialogAbout
from Nagstamon.qui.dialogs.acknowledge import DialogAcknowledge
from Nagstamon.qui.dialogs.action import DialogAction
from Nagstamon.qui.dialogs.authentication import DialogAuthentication
from Nagstamon.qui.dialogs.dialog import Dialog
from Nagstamon.qui.dialogs.downtime import DialogDowntime
from Nagstamon.qui.dialogs.server import DialogServer
from Nagstamon.qui.dialogs.server_missing import DialogServerMissing
from Nagstamon.qui.dialogs.settings import DialogSettings
from Nagstamon.qui.dialogs.submit import DialogSubmit
from Nagstamon.qui.dialogs.weblogin import DialogWebLogin
from Nagstamon.qui.helpers import hide_macos_dock_icon
from Nagstamon.qui.qt import (QObject,
                              Slot)


class Dialogs(QObject):
    """
    class for accessing all dialogs
    """
    about = None
    acknowledge = None
    action = None
    authentication = None
    downtime = None
    submit = None
    server = None
    server_missing = None
    settings = None
    weblogin = None
    windows = list()

    def initialize_dialog_settings(self, dialog):
        """
        initialize settings dialog
        """
        self.settings = dialog
        self.settings.initialize()
        self.windows.append(self.settings.window)

    def initialize_dialog_server(self, dialog):
        """
        initialize settings dialog
        """
        self.server = dialog
        self.server.initialize()
        self.windows.append(self.server.window)
        # check if special widgets have to be shown
        self.server.edited.connect(self.settings.toggle_zabbix_widgets)
        self.server.edited.connect(self.settings.toggle_op5monitor_widgets)
        self.server.edited.connect(self.settings.toggle_expire_time_widgets)

    def initialize_dialog_action(self, dialog):
        self.action = dialog
        self.action.initialize()
        self.windows.append(self.action.window)

    def initialize_dialog_acknowledge(self, dialog):
        self.acknowledge = dialog
        self.acknowledge.initialize()
        self.windows.append(self.acknowledge.window)
        self.acknowledge.window.button_change_defaults_acknowledge.clicked.connect(self.settings.show_defaults)
        self.acknowledge.window.button_change_defaults_acknowledge.clicked.connect(self.acknowledge.window.close)

    def initialize_dialog_downtime(self, dialog):
        self.downtime = dialog
        self.downtime.initialize()
        self.windows.append(self.downtime.window)
        self.downtime.window.button_change_defaults_downtime.clicked.connect(self.settings.show_defaults)
        self.downtime.window.button_change_defaults_downtime.clicked.connect(self.downtime.window.close)

    def initialize_dialog_submit(self, dialog):
        self.submit = dialog
        self.submit.initialize()
        self.windows.append(self.submit.window)

    def initialize_dialog_authentication(self, dialog):
        self.authentication = dialog
        self.authentication.initialize()
        self.windows.append(self.authentication.window)

    def initialize_dialog_server_missing(self, dialog):
        self.server_missing = dialog
        self.server_missing.initialize()
        self.windows.append(self.server_missing.window)
        # open server creation dialog
        self.server_missing.window.button_create_server.clicked.connect(self.settings.show_new_server)
        self.server_missing.window.button_enable_server.clicked.connect(self.settings.show)

    def initialize_dialog_about(self, dialog):
        self.about = dialog
        self.windows.append(self.about.window)

    def initialize_dialog_weblogin(self, dialog):
        """
        initialize web login dialog
        """
        self.weblogin = dialog
        self.weblogin.initialize()
        self.windows.append(self.weblogin.window)

    def get_shown_dialogs(self):
        """
        get a list of currently show dialog windows - needed for macOS hide dock icon stuff
        """
        return [x for x in self.windows if x.isVisible()]

    @Slot()
    def show_macos_dock_icon_if_necessary(self):
        """
        show macOS dock icon again if it is configured to be hidden
        was only necessary to show up to let dialog get keyboard focus
        """
        if OS == OS_MACOS and \
                conf.icon_in_systray and \
                conf.hide_macos_dock_icon:
            # if no window is shown already show dock icon
            if not len(self.get_shown_dialogs()):
                hide_macos_dock_icon(False)

    @Slot()
    def hide_macos_dock_icon_if_necessary(self):
        """
        hide macOS dock icon again if it is configured to be hidden
        was only necessary to show up to let the dialog get keyboard focus
        """
        if OS == OS_MACOS and \
                conf.icon_in_systray and \
                conf.hide_macos_dock_icon:
            # if no window is shown anymore hide dock icon
            if not len(self.get_shown_dialogs()):
                hide_macos_dock_icon(True)


dialogs = Dialogs()
dialogs.initialize_dialog_settings(DialogSettings())
dialogs.initialize_dialog_about(DialogAbout())
dialogs.initialize_dialog_acknowledge(DialogAcknowledge())
dialogs.initialize_dialog_action(DialogAction())
dialogs.initialize_dialog_downtime(DialogDowntime())
dialogs.initialize_dialog_submit(DialogSubmit())
dialogs.initialize_dialog_authentication(DialogAuthentication())
dialogs.initialize_dialog_server_missing(DialogServerMissing())
dialogs.initialize_dialog_server(DialogServer())
dialogs.initialize_dialog_weblogin(DialogWebLogin())

# signals and slots between dialogs
# settings -> server
dialogs.settings.server_created.connect(dialogs.server.new)
dialogs.settings.server_edited.connect(dialogs.server.edit)
dialogs.settings.server_copied.connect(dialogs.server.copy)
# settings -> action
dialogs.settings.action_created.connect(dialogs.action.new)
dialogs.settings.action_edited.connect(dialogs.action.edit)
dialogs.settings.action_copied.connect(dialogs.action.copy)
# action -> settings refresh_list
dialogs.action.edited_update_list.connect(dialogs.settings.refresh_list)
# server -> settings refresh_list
dialogs.server.edited_update_list.connect(dialogs.settings.refresh_list)
# servers and actions list update
dialogs.settings.update_list.connect(dialogs.settings.refresh_list)
# macOS dock icon fix
dialogs.action.check_macos_dock_icon_fix_show.connect(dialogs.show_macos_dock_icon_if_necessary)
dialogs.action.check_macos_dock_icon_fix_hide.connect(dialogs.hide_macos_dock_icon_if_necessary)
dialogs.authentication.check_macos_dock_icon_fix_show.connect(dialogs.show_macos_dock_icon_if_necessary)
dialogs.authentication.check_macos_dock_icon_fix_hide.connect(dialogs.hide_macos_dock_icon_if_necessary)
dialogs.authentication.check_macos_dock_icon_fix_show.connect(dialogs.show_macos_dock_icon_if_necessary)
dialogs.authentication.check_macos_dock_icon_fix_hide.connect(dialogs.hide_macos_dock_icon_if_necessary)
dialogs.server.check_macos_dock_icon_fix_show.connect(dialogs.show_macos_dock_icon_if_necessary)
dialogs.server.check_macos_dock_icon_fix_hide.connect(dialogs.hide_macos_dock_icon_if_necessary)


XK_Greek_ALPHAaccent = 0x7a1
XK_Greek_EPSILONaccent = 0x7a2
XK_Greek_ETAaccent = 0x7a3
XK_Greek_IOTAaccent = 0x7a4
XK_Greek_IOTAdiaeresis = 0x7a5
XK_Greek_OMICRONaccent = 0x7a7
XK_Greek_UPSILONaccent = 0x7a8
XK_Greek_UPSILONdieresis = 0x7a9
XK_Greek_OMEGAaccent = 0x7ab
XK_Greek_accentdieresis = 0x7ae
XK_Greek_horizbar = 0x7af
XK_Greek_alphaaccent = 0x7b1
XK_Greek_epsilonaccent = 0x7b2
XK_Greek_etaaccent = 0x7b3
XK_Greek_iotaaccent = 0x7b4
XK_Greek_iotadieresis = 0x7b5
XK_Greek_iotaaccentdieresis = 0x7b6
XK_Greek_omicronaccent = 0x7b7
XK_Greek_upsilonaccent = 0x7b8
XK_Greek_upsilondieresis = 0x7b9
XK_Greek_upsilonaccentdieresis = 0x7ba
XK_Greek_omegaaccent = 0x7bb
XK_Greek_ALPHA = 0x7c1
XK_Greek_BETA = 0x7c2
XK_Greek_GAMMA = 0x7c3
XK_Greek_DELTA = 0x7c4
XK_Greek_EPSILON = 0x7c5
XK_Greek_ZETA = 0x7c6
XK_Greek_ETA = 0x7c7
XK_Greek_THETA = 0x7c8
XK_Greek_IOTA = 0x7c9
XK_Greek_KAPPA = 0x7ca
XK_Greek_LAMDA = 0x7cb
XK_Greek_LAMBDA = 0x7cb
XK_Greek_MU = 0x7cc
XK_Greek_NU = 0x7cd
XK_Greek_XI = 0x7ce
XK_Greek_OMICRON = 0x7cf
XK_Greek_PI = 0x7d0
XK_Greek_RHO = 0x7d1
XK_Greek_SIGMA = 0x7d2
XK_Greek_TAU = 0x7d4
XK_Greek_UPSILON = 0x7d5
XK_Greek_PHI = 0x7d6
XK_Greek_CHI = 0x7d7
XK_Greek_PSI = 0x7d8
XK_Greek_OMEGA = 0x7d9
XK_Greek_alpha = 0x7e1
XK_Greek_beta = 0x7e2
XK_Greek_gamma = 0x7e3
XK_Greek_delta = 0x7e4
XK_Greek_epsilon = 0x7e5
XK_Greek_zeta = 0x7e6
XK_Greek_eta = 0x7e7
XK_Greek_theta = 0x7e8
XK_Greek_iota = 0x7e9
XK_Greek_kappa = 0x7ea
XK_Greek_lamda = 0x7eb
XK_Greek_lambda = 0x7eb
XK_Greek_mu = 0x7ec
XK_Greek_nu = 0x7ed
XK_Greek_xi = 0x7ee
XK_Greek_omicron = 0x7ef
XK_Greek_pi = 0x7f0
XK_Greek_rho = 0x7f1
XK_Greek_sigma = 0x7f2
XK_Greek_finalsmallsigma = 0x7f3
XK_Greek_tau = 0x7f4
XK_Greek_upsilon = 0x7f5
XK_Greek_phi = 0x7f6
XK_Greek_chi = 0x7f7
XK_Greek_psi = 0x7f8
XK_Greek_omega = 0x7f9
XK_Greek_switch = 0xFF7E



XK_hebrew_doublelowline = 0xcdf
XK_hebrew_aleph = 0xce0
XK_hebrew_bet = 0xce1
XK_hebrew_beth = 0xce1
XK_hebrew_gimel = 0xce2
XK_hebrew_gimmel = 0xce2
XK_hebrew_dalet = 0xce3
XK_hebrew_daleth = 0xce3
XK_hebrew_he = 0xce4
XK_hebrew_waw = 0xce5
XK_hebrew_zain = 0xce6
XK_hebrew_zayin = 0xce6
XK_hebrew_chet = 0xce7
XK_hebrew_het = 0xce7
XK_hebrew_tet = 0xce8
XK_hebrew_teth = 0xce8
XK_hebrew_yod = 0xce9
XK_hebrew_finalkaph = 0xcea
XK_hebrew_kaph = 0xceb
XK_hebrew_lamed = 0xcec
XK_hebrew_finalmem = 0xced
XK_hebrew_mem = 0xcee
XK_hebrew_finalnun = 0xcef
XK_hebrew_nun = 0xcf0
XK_hebrew_samech = 0xcf1
XK_hebrew_samekh = 0xcf1
XK_hebrew_ayin = 0xcf2
XK_hebrew_finalpe = 0xcf3
XK_hebrew_pe = 0xcf4
XK_hebrew_finalzade = 0xcf5
XK_hebrew_finalzadi = 0xcf5
XK_hebrew_zade = 0xcf6
XK_hebrew_zadi = 0xcf6
XK_hebrew_qoph = 0xcf7
XK_hebrew_kuf = 0xcf7
XK_hebrew_resh = 0xcf8
XK_hebrew_shin = 0xcf9
XK_hebrew_taw = 0xcfa
XK_hebrew_taf = 0xcfa
XK_Hebrew_switch = 0xFF7E



# encoding: utf-8

# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

"""Module servers"""

import urllib.request
import urllib.error
import urllib.parse

from collections import OrderedDict

# load all existing server types
from Nagstamon.servers.Nagios import NagiosServer
from Nagstamon.servers.Centreon import CentreonServer
from Nagstamon.servers.Icinga import IcingaServer
from Nagstamon.servers.IcingaWeb2 import IcingaWeb2Server
from Nagstamon.servers.IcingaDBWeb import IcingaDBWebServer
from Nagstamon.servers.IcingaDBWebNotifications import IcingaDBWebNotificationsServer
from Nagstamon.servers.Icinga2API import Icinga2APIServer
from Nagstamon.servers.Multisite import MultisiteServer
from Nagstamon.servers.op5Monitor import Op5MonitorServer
from Nagstamon.servers.Opsview import OpsviewServer
from Nagstamon.servers.Thruk import ThrukServer
from Nagstamon.servers.Zabbix import ZabbixServer
from Nagstamon.servers.ZabbixProblemBased import ZabbixProblemBasedServer
from Nagstamon.servers.Livestatus import LivestatusServer
from Nagstamon.servers.Zenoss import ZenossServer
from Nagstamon.servers.Monitos3 import Monitos3Server
from Nagstamon.servers.Monitos4x import Monitos4xServer
from Nagstamon.servers.SnagView3 import SnagViewServer
from Nagstamon.servers.Sensu import SensuServer
from Nagstamon.servers.SensuGo import SensuGoServer
from Nagstamon.servers.Prometheus import PrometheusServer
from Nagstamon.servers.Alertmanager import AlertmanagerServer
from Nagstamon.servers.LibreNMS import LibreNMSServer

from Nagstamon.config import conf

from Nagstamon.helpers import STATES


# dictionary for servers
servers = OrderedDict()

# contains dict with available server classes
# key is type of server, value is server class
# used for automatic config generation
# and holding this information in one place
SERVER_TYPES = OrderedDict()


def register_server(server):
    """
    Once new server class is created, should be registered with this function
    for being visible in config and accessible in application.
    """
    if server.TYPE not in SERVER_TYPES:
        SERVER_TYPES[server.TYPE] = server


def get_enabled_servers():
    """
    list of enabled servers which connections outside should be used to check
    """
    return [x for x in servers.values() if x.enabled is True]


def get_worst_status():
    """
    get the worst status of all servers
    """
    worst_status = 'UP'
    for server in get_enabled_servers():
        worst_status_current = server.get_worst_status_current()
        if STATES.index(worst_status_current) > STATES.index(worst_status):
            worst_status = worst_status_current
        del worst_status_current
    return worst_status


def get_status_count():
    """
    get all states of all servers and count them
    """
    state_count = {'UNKNOWN': 0,
                   'INFORMATION': 0,
                   'WARNING': 0,
                   'AVERAGE': 0,
                   'HIGH': 0,
                   'CRITICAL': 0,
                   'DISASTER': 0,
                   'UNREACHABLE': 0,
                   'DOWN': 0}
    for server in get_enabled_servers():
        state_count['UNKNOWN'] += server.unknown
        state_count['INFORMATION'] += server.information
        state_count['WARNING'] += server.warning
        state_count['AVERAGE'] += server.average
        state_count['HIGH'] += server.high
        state_count['CRITICAL'] += server.critical
        state_count['DISASTER'] += server.disaster
        state_count['UNREACHABLE'] += server.unreachable
        state_count['DOWN'] += server.down

    return state_count


def get_errors():
    """
    find out if any server has any error, used by statusbar error label
    """
    for server in get_enabled_servers():
        if server.has_error:
            return True
            break

    # return default value
    return False


def create_server(server=None):
    """
    create server from config
    """
    if server.type not in SERVER_TYPES:
        print(f'Server type not supported: {server.type}')
        return
    # give argument servername so CentreonServer could use it for initializing MD5 cache
    new_server = SERVER_TYPES[server.type](name=server.name)
    # apparently somewhat hacky but at the end works - might be used for others than Centreon as well
    if hasattr(new_server, 'ClassServerReal'):
        new_server = new_server.ClassServerReal(name=server.name)
    new_server.type = server.type
    new_server.enabled = server.enabled
    new_server.monitor_url = server.monitor_url
    new_server.monitor_cgi_url = server.monitor_cgi_url
    new_server.username = server.username
    new_server.password = server.password
    new_server.use_proxy = server.use_proxy
    new_server.use_proxy_from_os = server.use_proxy_from_os
    new_server.proxy_address = server.proxy_address
    new_server.proxy_username = server.proxy_username
    new_server.proxy_password = server.proxy_password
    new_server.authentication = server.authentication
    # The default timeout setting was increased to 30 seconds, but the old value of 10 seconds
    # will always be returned back coming from config file.
    # To avoid this mess it is now hardcoded to 30 seconds and the old value is not used anymore.
    #new_server.timeout = server.timeout
    new_server.timeout = 30

    # SSL/TLS
    new_server.ignore_cert = server.ignore_cert
    new_server.custom_cert_use = server.custom_cert_use
    new_server.custom_cert_ca_file = server.custom_cert_ca_file

    # ECP authentication
    new_server.idp_ecp_endpoint = server.idp_ecp_endpoint

    # if password is not to be saved ask for it at startup
    if (server.enabled is True and server.save_password is False and
            server.use_autologin is False):
        new_server.refresh_authentication = True

    # Special FX
    # Centreon
    new_server.use_autologin = server.use_autologin
    new_server.autologin_key = server.autologin_key

    # Icinga
    new_server.use_display_name_host = server.use_display_name_host
    new_server.use_display_name_service = server.use_display_name_service

    # IcingaWeb2
    new_server.no_cookie_auth = server.no_cookie_auth

    # IcingaDBWebNotifications
    new_server.notification_filter = server.notification_filter
    new_server.notification_lookback = server.notification_lookback

    # IcingaDBWeb custom filter
    new_server.custom_filter = server.custom_filter

    # Checkmk Multisite
    new_server.force_authuser = server.force_authuser
    new_server.checkmk_view_hosts = server.checkmk_view_hosts
    new_server.checkmk_view_services = server.checkmk_view_services

    # OP5 api filters
    new_server.host_filter = server.host_filter
    new_server.service_filter = server.service_filter

    # Opsview hashtag filter and can_change_only option
    new_server.hashtag_filter = server.hashtag_filter
    new_server.can_change_only = server.can_change_only

    # Zabbix
    new_server.use_description_name_service = server.use_description_name_service

    # Prometheus & Alertmanager
    new_server.alertmanager_filter = server.alertmanager_filter
    new_server.map_to_hostname = server.map_to_hostname
    new_server.map_to_servicename = server.map_to_servicename
    new_server.map_to_status_information = server.map_to_status_information
    new_server.map_to_ok = server.map_to_ok
    new_server.map_to_unknown = server.map_to_unknown
    new_server.map_to_warning = server.map_to_warning
    new_server.map_to_critical = server.map_to_critical
    new_server.map_to_down = server.map_to_down

    # LibreNMS
    new_server.treat_services_as_alerts = server.treat_services_as_alerts

    # Thruk
    new_server.disabled_backends = server.disabled_backends

    # server's individual preparations for HTTP connections (for example cookie creation)
    # is done in GetStatus() method of monitor
    if server.enabled is True:
        new_server.enabled = True

    # start with high thread counter so server update thread does not have to wait
    new_server.thread_counter = conf.update_interval_seconds

    # debug
    if conf.debug_mode is True:
        new_server.debug(server=server.name, debug="Created server.")

    return new_server


# moved registration process here because of circular dependencies
servers_list = [AlertmanagerServer,
                CentreonServer,
                IcingaServer,
                IcingaDBWebServer,
                IcingaDBWebNotificationsServer,
                IcingaWeb2Server,
                Icinga2APIServer,
                LibreNMSServer,
                LivestatusServer,
                Monitos3Server,
                Monitos4xServer,
                MultisiteServer,
                NagiosServer,
                Op5MonitorServer,
                OpsviewServer,
                PrometheusServer,
                SensuGoServer,
                SensuServer,
                SnagViewServer,
                ThrukServer,
                ZabbixProblemBasedServer,
                ZabbixServer,
                ZenossServer]

for server in servers_list:
    register_server(server)

# create servers
for server in conf.servers.values():
    created_server = create_server(server)
    if created_server is not None:
        servers[server.name] = created_server
        # for the next time no auth needed
        servers[server.name].refresh_authentication = False





# Xlib.X -- basic X constants
#
#    Copyright (C) 2000 Peter Liljenberg <petli@ctrl-c.liu.se>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,  USA

# Avoid overwriting None if doing "from Xlib.X import *"
NONE                = 0

ParentRelative      = 1  # background pixmap in CreateWindow
                         # and ChangeWindowAttributes

CopyFromParent      = 0  # border pixmap in CreateWindow
                         # and ChangeWindowAttributes
                         # special VisualID and special window
                         # class passed to CreateWindow

PointerWindow       = 0  # destination window in SendEvent
InputFocus          = 1  # destination window in SendEvent
PointerRoot         = 1  # focus window in SetInputFocus
AnyPropertyType     = 0  # special Atom, passed to GetProperty
AnyKey              = 0  # special Key Code, passed to GrabKey
AnyButton           = 0  # special Button Code, passed to GrabButton
AllTemporary        = 0  # special Resource ID passed to KillClient
CurrentTime         = 0  # special Time
NoSymbol            = 0  # special KeySym


#-----------------------------------------------------------------------
# Event masks:
#
NoEventMask               = 0
KeyPressMask              = (1<<0)
KeyReleaseMask            = (1<<1)
ButtonPressMask           = (1<<2)
ButtonReleaseMask         = (1<<3)
EnterWindowMask           = (1<<4)
LeaveWindowMask           = (1<<5)
PointerMotionMask         = (1<<6)
PointerMotionHintMask     = (1<<7)
Button1MotionMask         = (1<<8)
Button2MotionMask         = (1<<9)
Button3MotionMask         = (1<<10)
Button4MotionMask         = (1<<11)
Button5MotionMask         = (1<<12)
ButtonMotionMask          = (1<<13)
KeymapStateMask           = (1<<14)
ExposureMask              = (1<<15)
VisibilityChangeMask      = (1<<16)
StructureNotifyMask       = (1<<17)
ResizeRedirectMask        = (1<<18)
SubstructureNotifyMask    = (1<<19)
SubstructureRedirectMask  = (1<<20)
FocusChangeMask           = (1<<21)
PropertyChangeMask        = (1<<22)
ColormapChangeMask        = (1<<23)
OwnerGrabButtonMask       = (1<<24)

#-----------------------------------------------------------------------
# Event names:
#
# Used in "type" field in XEvent structures.  Not to be confused with event
# masks above.  They start from 2 because 0 and 1 are reserved in the
# protocol for errors and replies.
#
KeyPress               = 2
KeyRelease             = 3
ButtonPress            = 4
ButtonRelease          = 5
MotionNotify           = 6
EnterNotify            = 7
LeaveNotify            = 8
FocusIn                = 9
FocusOut               = 10
KeymapNotify           = 11
Expose                 = 12
GraphicsExpose         = 13
NoExpose               = 14
VisibilityNotify       = 15
CreateNotify           = 16
DestroyNotify          = 17
UnmapNotify            = 18
MapNotify              = 19
MapRequest             = 20
ReparentNotify         = 21
ConfigureNotify        = 22
ConfigureRequest       = 23
GravityNotify          = 24
ResizeRequest          = 25
CirculateNotify        = 26
CirculateRequest       = 27
PropertyNotify         = 28
SelectionClear         = 29
SelectionRequest       = 30
SelectionNotify        = 31
ColormapNotify         = 32
ClientMessage          = 33
MappingNotify          = 34
LASTEvent              = 35    # must be bigger than any event


#-----------------------------------------------------------------------
# Key masks:
#
# Used as modifiers to GrabButton and GrabKey, results of QueryPointer,
# state in various key-, mouse-, and button-related events.
#
ShiftMask = (1<<0)
LockMask = (1<<1)
ControlMask = (1<<2)
Mod1Mask = (1<<3)
Mod2Mask = (1<<4)
Mod3Mask = (1<<5)
Mod4Mask = (1<<6)
Mod5Mask = (1<<7)


#-----------------------------------------------------------------------
# Modifier names:
#
# Used to build a SetModifierMapping request or to read a
# GetModifierMapping request.  These correspond to the masks defined above.
#
ShiftMapIndex = 0
LockMapIndex = 1
ControlMapIndex = 2
Mod1MapIndex = 3
Mod2MapIndex = 4
Mod3MapIndex = 5
Mod4MapIndex = 6
Mod5MapIndex = 7

#-----------------------------------------------------------------------
# Button masks:
#
# Used in same manner as Key masks above. Not to be confused with button
# names below.  Note that 0 is already defined above as "AnyButton".
#
Button1Mask            = (1<<8)
Button2Mask            = (1<<9)
Button3Mask            = (1<<10)
Button4Mask            = (1<<11)
Button5Mask            = (1<<12)

AnyModifier            = (1<<15)  # used in GrabButton, GrabKey

#-----------------------------------------------------------------------
# Button names:
#
# Used as arguments to GrabButton and as detail in ButtonPress and
# ButtonRelease events.  Not to be confused with button masks above.
# Note that 0 is already defined above as "AnyButton".
#
Button1                = 1
Button2                = 2
Button3                = 3
Button4                = 4
Button5                = 5


#-----------------------------------------------------------------------
# XXX These still need documentation -- for now, read <X11/X.h>
#
NotifyNormal = 0
NotifyGrab = 1
NotifyUngrab = 2
NotifyWhileGrabbed = 3
NotifyHint = 1
NotifyAncestor = 0
NotifyVirtual = 1
NotifyInferior = 2
NotifyNonlinear = 3
NotifyNonlinearVirtual = 4
NotifyPointer = 5
NotifyPointerRoot = 6
NotifyDetailNone = 7
VisibilityUnobscured = 0
VisibilityPartiallyObscured = 1
VisibilityFullyObscured = 2
PlaceOnTop = 0
PlaceOnBottom = 1
FamilyInternet = 0
FamilyDECnet = 1
FamilyChaos = 2
PropertyNewValue = 0
PropertyDelete = 1
ColormapUninstalled = 0
ColormapInstalled = 1
GrabModeSync = 0
GrabModeAsync = 1
GrabSuccess = 0
AlreadyGrabbed = 1
GrabInvalidTime = 2
GrabNotViewable = 3
GrabFrozen = 4
AsyncPointer = 0
SyncPointer = 1
ReplayPointer = 2
AsyncKeyboard = 3
SyncKeyboard = 4
ReplayKeyboard = 5
AsyncBoth = 6
SyncBoth = 7
RevertToNone = 0
RevertToPointerRoot = PointerRoot
RevertToParent = 2
Success = 0
BadRequest = 1
BadValue = 2
BadWindow = 3
BadPixmap = 4
BadAtom = 5
BadCursor = 6
BadFont = 7
BadMatch = 8
BadDrawable = 9
BadAccess = 10
BadAlloc = 11
BadColor = 12
BadGC = 13
BadIDChoice = 14
BadName = 15
BadLength = 16
BadImplementation = 17
FirstExtensionError = 128
LastExtensionError = 255
InputOutput = 1
InputOnly = 2
CWBackPixmap = (1<<0)
CWBackPixel = (1<<1)
CWBorderPixmap = (1<<2)
CWBorderPixel = (1<<3)
CWBitGravity = (1<<4)
CWWinGravity = (1<<5)
CWBackingStore = (1<<6)
CWBackingPlanes = (1<<7)
CWBackingPixel = (1<<8)
CWOverrideRedirect = (1<<9)
CWSaveUnder = (1<<10)
CWEventMask = (1<<11)
CWDontPropagate = (1<<12)
CWColormap = (1<<13)
CWCursor = (1<<14)
CWX = (1<<0)
CWY = (1<<1)
CWWidth = (1<<2)
CWHeight = (1<<3)
CWBorderWidth = (1<<4)
CWSibling = (1<<5)
CWStackMode = (1<<6)
ForgetGravity = 0
NorthWestGravity = 1
NorthGravity = 2
NorthEastGravity = 3
WestGravity = 4
CenterGravity = 5
EastGravity = 6
SouthWestGravity = 7
SouthGravity = 8
SouthEastGravity = 9
StaticGravity = 10
UnmapGravity = 0
NotUseful = 0
WhenMapped = 1
Always = 2
IsUnmapped = 0
IsUnviewable = 1
IsViewable = 2
SetModeInsert = 0
SetModeDelete = 1
DestroyAll = 0
RetainPermanent = 1
RetainTemporary = 2
Above = 0
Below = 1
TopIf = 2
BottomIf = 3
Opposite = 4
RaiseLowest = 0
LowerHighest = 1
PropModeReplace = 0
PropModePrepend = 1
PropModeAppend = 2
GXclear = 0x0
GXand = 0x1
GXandReverse = 0x2
GXcopy = 0x3
GXandInverted = 0x4
GXnoop = 0x5
GXxor = 0x6
GXor = 0x7
GXnor = 0x8
GXequiv = 0x9
GXinvert = 0xa
GXorReverse = 0xb
GXcopyInverted = 0xc
GXorInverted = 0xd
GXnand = 0xe
GXset = 0xf
LineSolid = 0
LineOnOffDash = 1
LineDoubleDash = 2
CapNotLast = 0
CapButt = 1
CapRound = 2
CapProjecting = 3
JoinMiter = 0
JoinRound = 1
JoinBevel = 2
FillSolid = 0
FillTiled = 1
FillStippled = 2
FillOpaqueStippled = 3
EvenOddRule = 0
WindingRule = 1
ClipByChildren = 0
IncludeInferiors = 1
Unsorted = 0
YSorted = 1
YXSorted = 2
YXBanded = 3
CoordModeOrigin = 0
CoordModePrevious = 1
Complex = 0
Nonconvex = 1
Convex = 2
ArcChord = 0
ArcPieSlice = 1
GCFunction = (1<<0)
GCPlaneMask = (1<<1)
GCForeground = (1<<2)
GCBackground = (1<<3)
GCLineWidth = (1<<4)
GCLineStyle = (1<<5)
GCCapStyle = (1<<6)
GCJoinStyle = (1<<7)
GCFillStyle = (1<<8)
GCFillRule = (1<<9)
GCTile = (1<<10)
GCStipple = (1<<11)
GCTileStipXOrigin = (1<<12)
GCTileStipYOrigin = (1<<13)
GCFont = (1<<14)
GCSubwindowMode = (1<<15)
GCGraphicsExposures = (1<<16)
GCClipXOrigin = (1<<17)
GCClipYOrigin = (1<<18)
GCClipMask = (1<<19)
GCDashOffset = (1<<20)
GCDashList = (1<<21)
GCArcMode = (1<<22)
GCLastBit = 22
FontLeftToRight = 0
FontRightToLeft = 1
FontChange = 255
XYBitmap = 0
XYPixmap = 1
ZPixmap = 2
AllocNone = 0
AllocAll = 1
DoRed = (1<<0)
DoGreen = (1<<1)
DoBlue = (1<<2)
CursorShape = 0
TileShape = 1
StippleShape = 2
AutoRepeatModeOff = 0
AutoRepeatModeOn = 1
AutoRepeatModeDefault = 2
LedModeOff = 0
LedModeOn = 1
KBKeyClickPercent = (1<<0)
KBBellPercent = (1<<1)
KBBellPitch = (1<<2)
KBBellDuration = (1<<3)
KBLed = (1<<4)
KBLedMode = (1<<5)
KBKey = (1<<6)
KBAutoRepeatMode = (1<<7)
MappingSuccess = 0
MappingBusy = 1
MappingFailed = 2
MappingModifier = 0
MappingKeyboard = 1
MappingPointer = 2
DontPreferBlanking = 0
PreferBlanking = 1
DefaultBlanking = 2
DisableScreenSaver = 0
DisableScreenInterval = 0
DontAllowExposures = 0
AllowExposures = 1
DefaultExposures = 2
ScreenSaverReset = 0
ScreenSaverActive = 1
HostInsert = 0
HostDelete = 1
EnableAccess = 1
DisableAccess = 0
StaticGray = 0
GrayScale = 1
StaticColor = 2
PseudoColor = 3
TrueColor = 4
DirectColor = 5
LSBFirst = 0
MSBFirst = 1



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from Nagstamon.config import conf
from Nagstamon.qui.dialogs.dialog import Dialog
from Nagstamon.qui.qt import (Signal,
                              Slot)


class DialogDowntime(Dialog):
    """
    dialog for putting hosts/services into downtime
    """

    # send signal to get start and end of downtime asynchronously
    get_start_end = Signal(str, str)

    # signal to tell worker to commit downtime
    downtime = Signal(dict)

    # store host and service to be used for OK button evaluation
    server = None
    host_list = service_list = []

    def __init__(self):
        Dialog.__init__(self, 'dialog_downtime')

    def initialize(self, server=None, host=[], service=[]):
        # store server, host and service to be used for OK button evaluation
        self.server = server
        self.host_list = host
        self.service_list = service

        self.window.setWindowTitle('Downtime for host and service')
        text = ''

        for i in range(len(self.host_list)):
            if self.service_list[i] == "":
                text += f'Host <b>{self.host_list[i]}</b><br>'
            else:
                text += f'Service <b>{self.service_list[i]}</b> on host <b>{self.host_list[i]}</b><br>'

        self.window.input_label_description.setText(text)

        # default flags of monitor acknowledgement
        self.window.input_spinbox_duration_hours.setValue(int(conf.defaults_downtime_duration_hours))
        self.window.input_spinbox_duration_minutes.setValue(int(conf.defaults_downtime_duration_minutes))
        self.window.input_radiobutton_type_fixed.setChecked(conf.defaults_downtime_type_fixed)
        self.window.input_radiobutton_type_flexible.setChecked(conf.defaults_downtime_type_flexible)

        # hide/show downtime settings according to type
        self.window.input_radiobutton_type_fixed.clicked.connect(self.set_type_fixed)
        self.window.input_radiobutton_type_flexible.clicked.connect(self.set_type_flexible)

        # show or hide widgets for time settings
        if self.window.input_radiobutton_type_fixed.isChecked():
            self.set_type_fixed()
        else:
            self.set_type_flexible()

        # empty times at start, will be filled by set_start_end
        self.window.input_lineedit_start_time.setText('n/a')
        self.window.input_lineedit_end_time.setText('n/a')

        # default author + comment
        self.window.input_lineedit_comment.setText(conf.defaults_downtime_comment)
        self.window.input_lineedit_comment.setFocus()

        if self.server is not None:
            # at first initialization server is still None
            self.get_start_end.emit(self.server.name, self.host_list[0])

    def ok(self):
        """
        schedule downtime for miserable host/service
        """
        # type of downtime - fixed or flexible
        if self.window.input_radiobutton_type_fixed.isChecked() is True:
            fixed = 1
        else:
            fixed = 0

        for line_number in range(len(self.host_list)):
            service = self.service_list[line_number]
            host = self.host_list[line_number]

            self.downtime.emit({'server': self.server,
                                'host': host,
                                'service': service,
                                'author': self.server.username,
                                'comment': self.window.input_lineedit_comment.text(),
                                'fixed': fixed,
                                'start_time': self.window.input_lineedit_start_time.text(),
                                'end_time': self.window.input_lineedit_end_time.text(),
                                'hours': int(self.window.input_spinbox_duration_hours.value()),
                                'minutes': int(self.window.input_spinbox_duration_minutes.value())})
        # call close and macOS dock icon treatment from ancestor
        super().ok()

    @Slot(str, str)
    def set_start_end(self, start, end):
        """
        put values sent by worker into start and end fields
        """
        self.window.input_lineedit_start_time.setText(start)
        self.window.input_lineedit_end_time.setText(end)

    @Slot()
    def set_type_fixed(self):
        """
        enable/disable appropriate widgets if type is "Fixed"
        """
        self.window.label_duration.hide()
        self.window.label_duration_hours.hide()
        self.window.label_duration_minutes.hide()
        self.window.input_spinbox_duration_hours.hide()
        self.window.input_spinbox_duration_minutes.hide()

    @Slot()
    def set_type_flexible(self):
        """
        enable/disable appropriate widgets if type is "Flexible"
        """
        self.window.label_duration.show()
        self.window.label_duration_hours.show()
        self.window.label_duration_minutes.show()
        self.window.input_spinbox_duration_hours.show()
        self.window.input_spinbox_duration_minutes.show()



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from Nagstamon.qui.qt import QIcon


class QIconWithFilename(QIcon):
    """
    Extends QIcon with a filename attribute.

    This class allows assigning the filename of the icon file as an attribute to a QIcon object.

    Args:
        *args: Arguments passed to QIcon. If the first argument is a string, it is stored as the filename.
        **kwargs: Keyword arguments passed to QIcon.

    Attributes:
        filename (str): The filename of the icon, if specified during creation.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if type(args[0]) == str:
            self.filename = args[0]
